from unicon.eal.dialogs import Dialog

class AsaDialog:
    ssh_console_dialog = Dialog([
        ['continue connecting (yes/no)?', 'sendline(yes)', None, True, False],
        ['[P|p]assword:', 'sendline_ctx(password)', None, True, False],
        ['Last login:', None, None, False, False],
    ])

    ssh_vty_dialog = Dialog([
        ['continue connecting (yes/no)?', 'sendline(yes)', None, True, False],
        ['[P|p]assword:', 'sendline_ctx(password)', None, True, False],
        ['for a list of available commands.', None, None, False, False],
    ])

    download_dialog = Dialog([
        ['Address or name of remote host \[.*\]', 'sendline()', None, True, False],
        ['Source username \[.*\]', 'sendline()', None, True, False],
        ['Source filename \[.*\]', 'sendline()', None, True, False],
        ['Destination filename \[.*\]', 'sendline()', None, True, False],
        ['Do you want to over write\?', 'sendline_ctx(overwrite)', None, True, False],
        ['[.*>#] ', 'sendline()', None, False, False],
    ])



from unicon.statemachine import State, Path, StateMachine
from .statements import KpStatements
from .patterns import KpPatterns
from .dialogs import KpDialogs


class KpStateMachine(StateMachine):
    """An SSP class that restores all states."""
    def __init__(self, patterns):
        """Initializer of SspStateMachine."""

        self.patterns = patterns
        self.dialogs = KpDialogs(patterns)
        self.statements = KpStatements(patterns)
        super().__init__()

    def create(self):
        # Create States and their state patterns
        fxos_state = State('fxos_state', self.patterns.prompt.fxos_prompt)
        fireos_state = State('fireos_state', self.patterns.prompt.fireos_prompt)
        expert_state = State('expert_state', self.patterns.prompt.expert_prompt)
        sudo_state = State('sudo_state', self.patterns.prompt.sudo_prompt)
        rommon_state = State('rommon_state', self.patterns.prompt.rommon_prompt)
        local_mgmt_state = State('local_mgmt_state', self.patterns.prompt.local_mgmt_prompt)

        # lina cli states
        enable_state = State('enable_state', self.patterns.prompt.enable_prompt)
        disable_state = State('disable_state', self.patterns.prompt.disable_prompt)
        config_state = State('config_state', self.patterns.prompt.config_prompt)

        # Add our states to the state machine
        self.add_state(fxos_state)
        self.add_state(fireos_state)
        self.add_state(expert_state)
        self.add_state(sudo_state)
        self.add_state(rommon_state)
        self.add_state(local_mgmt_state)
        self.add_state(enable_state)
        self.add_state(disable_state)
        self.add_state(config_state)

        # Create paths for switching between states
        fxos_to_ftd = Path(fxos_state, fireos_state, "connect ftd", None)
        fxos_to_local_mgmt = Path(fxos_state, local_mgmt_state, "connect local-mgmt", None)
        ftd_to_expert = Path(fireos_state, expert_state, "expert", None)
        ftd_expert_to_sudo = Path(expert_state, sudo_state, "sudo su -", self.dialogs.d_expert_to_sudo)
        ftd_sudo_to_expert = Path(sudo_state, expert_state, "exit", None)
        expert_to_ftd = Path(expert_state, fireos_state, "exit", None)
        local_mgmt_to_fxos = Path(local_mgmt_state, fxos_state, "exit", None)
        # there are two paths from fireos_state to fxos_state.
        # 1. if on console, type "exit". if you type "connect fxos" you will be prompted to use "exit".
        # 2. if on mgmt vty, type "connect fxos". if you type "exit", you will lose the vty.
        # the following dialog tries "connect fxos" first, if prompted, uses "exit". it should work on both.
        ftd_to_fxos = Path(fireos_state, fxos_state, "connect fxos", self.dialogs.d_ftd_to_fxos)

        # Crete lina cli paths
        fireos_to_disable_state = Path(fireos_state, disable_state,
                                       'system support diagnostic-cli',
                                       None)
        disable_to_enable_state = Path(disable_state, enable_state, 'en',
                                       self.dialogs.ftd_dialogs.disable_to_enable)
        enable_to_disable_state = Path(enable_state, disable_state, "disable", None)

        disable_to_fireos_path = Path(disable_state, fireos_state, '\001'+'d'+'exit',
                                      None)

        enable_to_config_path = Path(enable_state, config_state, 'conf t', None)
        config_to_enable_path = Path(config_state, enable_state, 'end', None)

        # Add paths to the State Machine
        self.add_path(fxos_to_ftd)
        self.add_path(ftd_to_fxos)
        self.add_path(fxos_to_local_mgmt)
        self.add_path(local_mgmt_to_fxos)
        self.add_path(ftd_to_expert)
        self.add_path(expert_to_ftd)
        self.add_path(ftd_expert_to_sudo)
        self.add_path(ftd_sudo_to_expert)
        self.add_path(fireos_to_disable_state)
        self.add_path(disable_to_fireos_path)
        self.add_path(disable_to_enable_state)
        self.add_path(enable_to_disable_state)
        self.add_path(enable_to_config_path)
        self.add_path(config_to_enable_path)

        # after inactivity timer, it will go back to prelogin:
        self.add_default_statements([self.statements.login_incorrect, self.statements.login_password])



"""Library to load new images on a FMC Series3.

We require a file server to be co-located with the device. This file
server has both the img files mounted.  Image downloading is via http.

The file server is a PXE server in each site.

"""
import logging

import datetime

logger = logging.getLogger(__name__)

from unicon.eal.dialogs import Dialog
from ...series3.actions.webserver import Webserver
from ...fmc.actions import Fmc, FmcLine
from .constants import Series3Constants

try:
    from kick.graphite.graphite import publish_kick_metric
except ImportError:
    from kick.metrics.metrics import publish_kick_metric

KICK_EXTERNAL = False

try:
    from kick.file_servers.file_servers import *
except ImportError:
    KICK_EXTERNAL = True
    pass

try:
    from kick.kick_constants import KickConsts
except ImportError:
    from kick.miscellaneous.credentials import KickConsts


DUT_DIR_TMP = '/tmp'
DUT_DIR_BOOT = '/boot'
DUT_DIR_ETC = '/etc'
DUT_LILO_TMPL_FILE = 'lilo.conf'
DUT_LILO_FILE = 'lilo.conf.atf'
DUT_LILO_TMPL_PATH = '{}/{}'.format(DUT_DIR_ETC, DUT_LILO_TMPL_FILE)
DUT_LILO_TMP_PATH = '{}/{}'.format(DUT_DIR_TMP, DUT_LILO_FILE)
DUT_LILO_PATH = '{}/{}'.format(DUT_DIR_ETC, DUT_LILO_FILE)
MAX_TIME_INSTALL = 7200
DEFAULT_TIMEOUT = 10
MAX_RETRY_COUNT = 3
SSH_DEFAULT_TIMEOUT = 90


class Series3Fmc(Fmc):
    def __init__(self, hostname='firepower', login_username='admin',
                 login_password=KickConsts.DEFAULT_PASSWORD,
                 sudo_password=KickConsts.DEFAULT_PASSWORD, *args,
                 **kwargs):
        """Constructor for FMC.

        :param hostname: hostname of the device or fqdn
        :param login_username: user credentials
        :param login_password: device login password with user name 'admin'
        :param sudo_password: device sudo password for 'root'

        """

        super().__init__(hostname, login_username,
                         login_password,
                         sudo_password, *args, **kwargs)
        publish_kick_metric('device.series3fmc.init', 1)

        self.line_class = Series3FmcLine

    def ssh_vty(self, ip, port, username='admin', password='Admin123', timeout=None,
                line_type='ssh', rsa_key=None):
        """
        Set up an ssh connection to device's interface.

        This goes into device's ip address, not console.

         :param ip: ip address of terminal server
         :param port: port of device on terminal server
         :param username: usually "admin"
         :param password: usually "Admin123"
         :param timeout: in seconds
         :param line_type: ssh line type
         :param rsa_key: identity file (full path)
         :return: a line object (where users can call execute(), for example)

         """
        publish_kick_metric('device.series3fmc.ssh_vty', 1)

        from .constants import Series3Constants
        Series3Constants.uut_ssh_ip = ip
        Series3Constants.uut_ssh_port = port

        ssh_line = super().ssh_vty(ip=ip, port=port, username=username, password=password,
                                   timeout=timeout, line_type=line_type, rsa_key=rsa_key)

        return ssh_line


class Series3FmcLine(FmcLine):
    def __init__(self, spawn, sm, type, timeout=None):
        """Constructor for series3fmc line A

        :param spawn: spawn of the line
        :param sm: object of statemachine
        :param type: line type, e.g. 'ssh', 'telnet'
        :return: None

        """
        self.spawn_id = spawn
        self.sm = sm
        self.type = type
        self.set_default_timeout(DEFAULT_TIMEOUT)

        try:
            sm.go_to('any', spawn, timeout=timeout)
            if self.sm.current_state == 'prelogin_state':
                self.sendline(self.sm.patterns.login_username)
                self.spawn_id.expect('[p|P]assword: ')
                self.sendline(self.sm.patterns.login_password)
                self.go_to('any')
                if self.sm.current_state == 'admin_state':
                    self.sm.remove_path('prelogin_to_fireos_path')
                    self.sm.remove_path('fireos_to_prelogin_path')
                elif self.sm.current_state == 'fireos_state':
                    self.sm.remove_path('admin_to_prelogin')
                    self.sm.remove_path('prelogin_to_admin')
                else:
                    pass
        except:
            # Catch the case fmc is back from installation
            # Does not have the stable state
            pass
    # the following functions are for device baselining. This only works on
    # a ssh connection.
    def generate_config_file(self,private_iso=False):
        """Compose the config file. Copy the config file to
        self.http_dir_for_build on Webserver. Create symbolic link to the iso
        image on Webserver.
        :param private_iso: Flag for private iso configuration

        :return: None

        """

        server = Webserver(hostname=self.scp_hostname,
                           login_username=self.scp_username,
                           login_password=self.scp_password)
        dev = server.ssh_vty(self.scp_server)
        output = dev.execute('mkdir -p -m 775 {}'.format(self.http_dir_for_build))
        result = re.search('File exists', output)
        if result is not None:
            logger.info('File exists and it will be deleted ...')
            dev.execute('rm -rf {}'.format(self.http_dir_for_build))
            dev.execute('mkdir -p -m 775 {}'.format(self.http_dir_for_build))
        dev.execute('ln -s {}/{} {}'.format(self.http_dir_root,
                                            self.iso_image_path, self.http_dir_for_build))
        config_file_content = """TRANS=httpsrv
SRV={}
IPCONF=dhcp
REPLACE_TFTP_BOOT_FILE=Yes
TFTPSERVER={}
FN={}
TARGET_ARCH={}
""".format(self.http_server,
           self.http_server,
           self.iso_image_path,
           self.arch)
        dev.execute('echo "{}" > {}'.format(config_file_content, self.config_path))
        output_ls = dev.execute('ls {}'.format(self.config_path))
        logger.info(output_ls)
        output_cat = dev.execute('cat {}'.format(self.config_path))
        logger.info(output_cat)
        if private_iso:
            output = dev.execute('wget -O - http://{}:{}/{}'.format(self.scp_server, self.http_port, self.config_file_url))
            assert 'TRANS=httpsrv' in output, \
                    '\nFailed to access the boot configuration file.\nVerify http_dir_tmp is correct.\nhttp://{}:{}/{}'. \
                    format(self.scp_server, self.http_port, self.config_file_url)
        else:
            output = dev.execute(
                'wget -O - http://{}:{}/{}'.format(self.http_server, self.http_port, self.config_file_url))
            assert 'TRANS=httpsrv' in output, \
                '\nFailed to access the boot configuration file.\nVerify http_dir_tmp is correct.\nhttp://{}:{}/{}'. \
                    format(self.http_server, self.http_port, self.config_file_url)
        dev.disconnect()

    def scp_file_to_local(self, local_dir, sftp_ip, sftp_port,
                          username, password, remote_path):
        """Scp a remote file to DUT.

        :param local_dir: local directory as the destination directory on DUT
        :param sftp_ip: remote sftp server ip address
        :param sftp_port: remote sftp server port
        :param username: username to login to sftp server
        :param password: password to login to sftp server
        :param remote_path: file path with filename on remote sftp server
        :return: None

        """

        filename = remote_path.split('/')[-1]
        self.execute('rm -f {}/{}'.format(local_dir, filename))
        self.spawn_id.sendline('scp -P {} {}@{}:{} {}'.format(sftp_port,
                                                              username, sftp_ip, remote_path, local_dir))
        current_prompt = self.sm.get_state(self.sm.current_state).pattern
        d1 = Dialog([
            ['continue connecting (yes/no)?', 'sendline(yes)', None, True, False],
            ['[p|P]assword:', 'sendline({})'.format(password), None, True, False],
            [current_prompt, 'sendline()', None, False, False],
        ])
        d1.process(self.spawn_id, timeout=300)
        self.execute('pwd', timeout=180)
        self.execute('ls -ltr {}/{}'.format(local_dir, filename),timeout=120)
        output = self.execute('ls -ltr {}/{}'.format(local_dir, filename),
                              timeout=20)
        logger.info(output)
        full_filename = output.split('/')[-1]
        logger.info('filename={}, full_filename={}'.format(filename,
                                                           full_filename))
        assert ('No such file or directory' not in output) and \
               (filename[:-1] in full_filename), \
            'Failed to copy {}'.format(filename)
        return full_filename

    def copy_bz_usb_files_setup_lilo_to_install(self,private_iso=False):
        """Scp bz and usb image files from remote sftp server to DUT. Configure
        lilo file on DUT. Execute run_lilo_and_depmod. Mount usb. Reboot to
        install.
        :param private_iso: Flag for private iso lilo configuration

        :return: None

        """

        # Scp bz and usb files
        self.go_to('sudo_state')

        full_bz_filename = self.scp_file_to_local(local_dir=DUT_DIR_TMP,
                                                  sftp_ip=self.scp_server, sftp_port=self.scp_port,
                                                  username=self.scp_username, password=self.scp_password,
                                                  remote_path=self.bz_image_path)
        logger.info('=== Extract bz_filename from {}: {}'.format(self.bz_image_path,
                                                                 full_bz_filename))
        full_usb_filename = self.scp_file_to_local(local_dir=DUT_DIR_TMP,
                                                   sftp_ip=self.scp_server, sftp_port=self.scp_port,
                                                   username=self.scp_username, password=self.scp_password,
                                                   remote_path=self.usb_image_path)
        logger.info('=== Extract usb_filename from {}: '
                    '{}'.format(self.usb_image_path,
                                full_usb_filename))
        cmd = 'mv {}/{} {}'.format(DUT_DIR_TMP,
                                   full_bz_filename,
                                   DUT_DIR_BOOT)
        self.execute_only(cmd=cmd)

        cmd = 'mv {}/{} {}'.format(DUT_DIR_TMP,
                                   full_usb_filename,
                                   DUT_DIR_BOOT)
        self.execute_only(cmd=cmd)

        # Generate lilo file
        cmd = "cat {} | perl -pe 's/default=[^\n]+/default=INSTALL/g'" \
              " > {}".format(DUT_LILO_TMPL_PATH, DUT_LILO_TMP_PATH)
        self.execute_only(cmd=cmd)
        if private_iso:

            lilo_append_lines = """
    image=/boot/{}
          label=INSTALL
          read-only
          initrd=/boot/{}
          append="root=/dev/ram0 rw PRESERVE_DATA=Yes choicedevice=eth0 INTEGCONF={}/{}"
    """.format(full_bz_filename, full_usb_filename, self.scp_server, self.config_file_url)
            cmd = '''echo '{}' >> {}'''.format(lilo_append_lines,
                                               DUT_LILO_TMP_PATH)
        else:
            lilo_append_lines = """
            image=/boot/{}
                  label=INSTALL
                  read-only
                  initrd=/boot/{}
                  append="root=/dev/ram0 rw PRESERVE_DATA=Yes choicedevice=eth0 INTEGCONF={}/{}"
            """.format(full_bz_filename, full_usb_filename, self.http_server, self.config_file_url)
            cmd = '''echo '{}' >> {}'''.format(lilo_append_lines,
                                               DUT_LILO_TMP_PATH)

        self.execute_only(cmd=cmd)

        self.execute('cat {}'.format(DUT_LILO_TMP_PATH))
        self.execute('mv {} {}'.format(DUT_LILO_TMP_PATH, DUT_LILO_PATH))
        logger.info('=== Generated lilo file [{}]\n{}\n'.format(DUT_LILO_PATH,
                                                                lilo_append_lines))
        logger.info('=== run_lilo_and_depmod')
        cmd = '/bin/bash -c "source /etc/lilo.d/functions.lilo; ' \
              'run_lilo_and_depmod"'
        self.execute_only(cmd=cmd, timeout=120)

        self.execute('mount LABEL=SFKEYFOB /usb', 60)
        self.execute('lilo -P ignore -C /etc/lilo.conf.atf', 60)
        self.execute('lilo -q 2>&1', 60)
        logger.info('=== System will reboot!')
        self.spawn_id.sendline('reboot')

    def confirm_device_rebooted(self):
        """Confirm device is rebooting by executing command pwd.

        If device is not reachable, exit. Wait until device is not
        reachable.

        """

        count = 0
        wait_time = 60
        total_count = 20
        while count < total_count:
            try:
                self.execute('pwd', 5)
                time.sleep(wait_time)
                count += 1
                logger.info('=== Wait {} seconds for device to reboot ...'.format(wait_time))
                time.sleep(wait_time)
            except:
                logger.info('=== The Appliance at {} is Rebooting.'.format(self.uut_ip))
                self.disconnect()
                break
        if count >= total_count:
            msg = '=== The appliance {} does not appear to have gone down' \
                  ' for reboot. Installer has failed'.format(self.uut_ip)
            logger.info(msg)
            raise RuntimeError(msg)

    def reconnect_and_process_dialog(self, device_reconnect_count, wait_time,
                                     dialog_timeout, start_time, dialog_name, dialog,initial_string="",initial_password="Admin123"):
        """Reconnect to the device. Process the dialog. Device could reboot
        multiple times. Wait until the device can be accessed and dialog is
        processed.

        :param device_reconnect_count: initial device reconnect count
        :param wait_time: wait time in each loop
        :param dialog_timeout: dialog timeout
        :param start_time: installation start time used to calculate elapsed_time
        :param dialog_name: dialog name used for logging info
        :param dialog: object of Dialog to be processed for device configuration
        :param initial_string: initial string sent to device to trigger dialog
        :param initial_password: initial password after iso is installed, the default value is "Admin123"
        :return device_reconnect_count: incremented device reconnect count

        """

        count = 0
        total_count = 20
        now = datetime.datetime.now()
        elapsed_time = (now - start_time).total_seconds()
        if elapsed_time < self.installation_timeout:
            logger.info('=== Sleep for {} seconds'.format(wait_time))
        while count < total_count:
            try:
                logger.info('=== Begin: {}'.format(dialog_name))
                time.sleep(wait_time)

                #initial_password = self.sm.patterns.default_password

                tmp_fmc = self.fmc
                tmp_uut_ip = self.uut_ip
                tmp_version_build = self.version_build

                conn = self.fmc.ssh_vty(ip=Series3Constants.uut_ssh_ip,
                                        port=Series3Constants.uut_ssh_port,
                                        password=initial_password,
                                        timeout=self.ssh_timeout)

                self.spawn_id = conn.spawn_id
                self.fmc = tmp_fmc
                self.uut_ip = tmp_uut_ip
                self.version_build = tmp_version_build

                device_reconnect_count += 1
                logger.info('Device reconnected, {} times'.format(device_reconnect_count))
                self.go_to('admin_state')
                if initial_string is not "":
                    self.spawn_id.sendline(initial_string)
                self.spawn_id.sendline()
                dialog.process(self.spawn_id, timeout=dialog_timeout)
                logger.info('=== End: {}'.format(dialog_name))
                time.sleep(5)
                if conn:
                    conn.disconnect()
                return device_reconnect_count
            except:
                now = datetime.datetime.now()
                elapsed_time = (now - start_time).total_seconds()
                logger.info('=== Device is not ready, elapsed_time={}, '
                            'wait for {} seconds ...'.format(elapsed_time, wait_time))
                pass

    def poll_device_and_configure(self,initial_password):
        """Poll device: wait for device to be up.
        :param initial_password: Password with which device will come up after installation
        :return: None

        """
        httpsd = '/usr/bin/httpsd -D FOREGROUND'
        d1 = Dialog([
            [httpsd, None, None, False, False],
            ])

        start_time = datetime.datetime.now()
        self.reconnect_and_process_dialog(
            device_reconnect_count=0,
            wait_time=600,
            dialog_timeout=30,
            start_time=start_time,
            initial_password= initial_password,
            dialog_name='wait for device to be ready',
            dialog=d1,
            initial_string='ps aux | grep www')

    def series3fmc_baseline(self, fmc, http_server, scp_server, scp_port,
                            scp_username, scp_password, scp_hostname,
                            version_build, iso_image_path,
                            uut_ip, uut_netmask, uut_gateway, dns_server,
                            hostname='firepower',
                            search_domains='cisco.com',
                            http_dir_root='/var/www',
                            http_dir_tmp='iso/tmp_dev',
                            scp_dir_root='/nfs/netboot/ims/Release',
                            bz_image='bzImage*',
                            usb_image='usb-ramdisk*',
                            arch='x86_64',
                            console=False,
                            uut_ip6=None, uut_prefix=None, uut_gateway6=None,
                            timeout=None, change_password=True, http_port='80',ipv4_mode='static',ipv6_mode='',ntp_servers=''):
        """Install FMC iso image on Series 3 device via http.

        :param fmc: object of Series3fmc for reconnecting the device
        :param http_server: HTTP Server IP Address, e.g. '10.89.23.80'
                            must use the same http server as the scp server
        :param http_port: HTTP server port, e.g. '80'
        :param scp_server: SCP Server IP Address, e.g. '10.89.23.80'
        :param scp_port: SCP Server Port, e.g. '22'
        :param scp_username: SCP Username
        :param scp_password: SCP Password
        :param scp_hostname: SCP Hostname
        :param version_build: e.g. '6.2.1-1366'
        :param iso_image_path: FMC image path on HTTP server
                                e.g. '/netboot/ims/Development/6.2.1-1366/iso/Sourcefire_Defense_Center_S3-6.2.1-1366-Restore.iso'
        :param uut_ip: Device IP Address
        :param uut_netmask: Device Netmask
        :param uut_gateway: Device Gateway
        :param dns_server: DNS server
        :param hostname: hostname to be set, e.g. 'BATIT-3D8130-1-AST'
        :param search_domains: search domains delimited by comma,
                                defaulted to 'cisco.com'
        :param http_dir_root: absolute path of the http root
                                e.g. '/var/www'
        :param http_dir_tmp: the temp dir under http_dir_root for version_build
                            and the config file, e.g. 'iso/tmp_dev'
        :param scp_dir_root: absolute path of the dir for version_build
                            e.g. '/nfs/netboot/ims/Release' or '/nfs/netboot/ims/Development'
        :param bz_image: bz image has to be installed on DUT, which is copied
                        from the scp_server, defaulted to 'bzImage*'
        :param usb_image: usb image has to be installed on DUT, which is copied
                            from the scp_server, defaulted to 'usb-ramdisk*'
        :param arch: arch type, defaulted to 'x86_64'
        :param console: set to True if the device is accessed through its console;
                        defaulted to False
        :param uut_ip6: Device IPv6 Address
        :param uut_prefix: Device IPv6 Prefix
        :param uut_gateway6: Device IPv6 Gateway
        :param timeout: in seconds; time to wait for device to boot with the new image;
                        if not provided, default baseline time is 7200s
        :param change_password: Flag to change the password after baseline
        :return: None

        """
        publish_kick_metric('device.series3fmc.baseline', 1)
        # set baseline timeout
        self.set_installation_timeouts(timeout)
        # Define attributes and variables
        self.fmc = fmc
        self.http_server = http_server
        self.http_port = http_port
        self.scp_server = scp_server
        self.scp_port = scp_port
        self.scp_username = scp_username
        self.scp_password = scp_password
        self.scp_hostname = scp_hostname
        self.version_build = version_build
        self.iso_image_path = iso_image_path
        self.iso_file = iso_image_path.split('/')[-1]
        self.uut_ip = uut_ip
        self.uut_netmask = uut_netmask
        self.uut_gateway = uut_gateway
        self.uut_ip6 = uut_ip6
        self.uut_prefix = uut_prefix
        self.uut_gateway6 = uut_gateway6
        self.dns_server = dns_server
        self.hostname = hostname
        self.search_domains = search_domains
        self.http_dir_root = http_dir_root
        self.http_dir_tmp = http_dir_tmp
        self.http_dir_for_build = '{}/{}/{}'.format(self.http_dir_root,
                                                    self.http_dir_tmp, self.version_build)
        self.config_file = '{}.config'.format(self.iso_file.strip('.iso'))
        self.config_path = '{}/{}'.format(self.http_dir_for_build, self.config_file)
        self.config_file_url = '{}/{}/{}'.format(self.http_dir_tmp, self.version_build, self.config_file)
        self.scp_dir_root = scp_dir_root
        self.bz_image = bz_image
        self.usb_image = usb_image
        self.arch = arch
        self.bz_image_path = '{}/{}/os/{}/boot/{}'.format(self.scp_dir_root,
                                                          self.version_build,
                                                          self.arch, self.bz_image)
        self.usb_image_path = '{}/{}/os/{}/ramdisks/{}'.format(self.scp_dir_root,
                                                               self.version_build,
                                                               self.arch, self.usb_image)
        self.console = console


        logger.info('=== Generate config file on the http server ...')
        self.generate_config_file()
        logger.info('=== Copy bz and usb files to dut ...')
        logger.info('=== Generate lilo file on dut ...')
        self.go_to('sudo_state')
        self.copy_bz_usb_files_setup_lilo_to_install()

        if not self.console:
            logger.info('=== Confirm device is rebooted ...')
            self.confirm_device_rebooted()
            logger.info('=== Wait for device to be up and configure device ...')
            self.poll_device_and_configure(initial_password=self.sm.patterns.default_password)
            conn = self.fmc.ssh_vty(ip=Series3Constants.uut_ssh_ip,
                                    port=Series3Constants.uut_ssh_port,
                                    password=self.sm.patterns.default_password,
                                    timeout=self.ssh_timeout)
            self.spawn_id = conn.spawn_id
        else:
            logger.info('=== Console connection available, wait for login prompt ...')
            d = Dialog([
                ['login:', 'sendline({})'.format(self.sm.patterns.login_username),
                 None, True, False],
                ['Password:', 'sendline({})'.format(self.sm.patterns.default_password),
                 None, False, False],
            ])
            d.process(self.spawn_id, timeout=self.installation_timeout)
            self._first_login()
            self.configuration_wizard(ipv4_mode=ipv4_mode, ipv6_mode=ipv6_mode,
                                      ipv4=uut_ip, ipv4_netmask=uut_netmask,
                                      ipv4_gateway=uut_gateway, ipv6=uut_ip6,
                                      ipv6_prefix=uut_prefix, ipv6_gateway=uut_gateway6,
                                      dns_servers=dns_server, search_domains=search_domains,
                                      ntp_servers=ntp_servers)
        self.go_to('any')
        self.configure_network(uut_gateway, uut_ip, uut_netmask,
                               uut_ip6, uut_prefix, uut_gateway6,
                               change_password)
        logger.info('=== Fully installed.')
        logger.info('=== Validate version installed ...')
        self.validate_version(self.iso_file)

        logger.info('Installation completed successfully.')

    def baseline_by_branch_and_version(self, fmc, site, branch, version,
                                       uut_ip, uut_netmask, uut_gateway,
                                       dns_server, iso_file_type='Restore',
                                       uut_ip6=None, uut_prefix=None, uut_gateway6=None,
                                       change_password=True, timeout=None, serverIp='', tftpPrefix='',
                                       scpPrefix='', docs='', pxeUsername='', pxePassword='', pxeHostname='',
                                       scpDirRoot='', httpDirRoot='', httpDirTmp='', **kwargs):
        """Baseline series3 fmc by branch and version using PXE servers.
        Look for needed files on devit-engfs, copy them to the local kick server
        and use them to baseline the device.

        :param fmc: object of Series3fmc for reconnecting the device
        :param site: e.g. 'ful', 'ast', 'bgl'
        :param branch: branch name, e.g. 'Release', 'Feature'
        :param version: software build-version, e,g, 6.2.3-623
        :param uut_ip: Device IP Address
        :param uut_netmask: Device Netmask
        :param uut_gateway: Device Gateway
        :param dns_server: DNS server
        :param iso_file_type: 'Autotest' or 'Restore'; defaulted to 'Restore'
        :param uut_ip6: Device IPv6 Address
        :param uut_prefix: Device IPv6 Prefix
        :param uut_gateway6: Device IPv6 Gateway
        :param timeout: in seconds; time to wait for device to boot with the new image;
                        if not provided, default baseline time is 7200s
        :param change_password: Flag to change the password after baseline
        :param \**kwargs:
        :Keyword Arguments, any of below optional parameters:
        :param hostname: hostname to be set, defaulted to 'firepower'
        :param search_domains: search domains delimited by comma,
                                defaulted to 'cisco.com'
        :param detection_mode: defaulted to 'Inline',
                                can be: 'Passive','Inline','Access Control','Network Discovery'
        :param http_dir_root: absolute path of the http root
                                e.g. '/tftpboot/asa/'
        :param http_dir_tmp: the temp dir under http_dir_root for version_build
                            and the config file, e.g. '/cache/temp_config'
        :param scp_dir_root: absolute path of the dir for version_build
                            e.g. '/tftpboot/asa/cache/Release' or '/tftpboot/asa/cache/Development'
        :param bz_image: bz image has to be installed on DUT, which is copied
                        from the scp_server, defaulted to 'bzImage*'
        :param usb_image: usb image has to be installed on DUT, which is copied
                            from the scp_server, defaulted to 'usb-ramdisk*'
        :param arch: arch type, defaulted to 'x86_64'
        :param console: set to True if the device is accessed through its console;
                        defaulted to False
        :return: None
        """

        kwargs['fmc'] = fmc
        kwargs['uut_ip'] = uut_ip
        kwargs['uut_netmask'] = uut_netmask
        kwargs['uut_gateway'] = uut_gateway
        kwargs['uut_ip6'] = uut_ip6
        kwargs['uut_prefix'] = uut_prefix
        kwargs['uut_gateway6'] = uut_gateway6
        kwargs['change_password'] = change_password
        kwargs['timeout'] = timeout
        kwargs['dns_server'] = dns_server
        arch = kwargs.get('arch', 'x86_64')

        if KICK_EXTERNAL:
            server_ip = serverIp
            tftp_prefix = tftpPrefix
            scp_prefix = scpPrefix
            files = docs
        else:
            server_ip, tftp_prefix, scp_prefix, files = prepare_installation_files(site, 's3fmc', branch, version,
                                                                                   iso=iso_file_type, arch_type=arch)
        if not files[0]:
            raise Exception('Series3 FMC iso file not found on server')

        logger.debug('=== server {}, tftp prefix {}, scp prefix {}'.format(server_ip, tftp_prefix, scp_prefix))
        iso_file_name = '{}/{}'.format(tftp_prefix[len('asa'):], files[0])
        kwargs['http_server'] = server_ip
        kwargs['scp_server'] = server_ip
        kwargs['scp_port'] = '22'
        if KICK_EXTERNAL:
            kwargs['scp_username'] = pxeUsername
            kwargs['scp_password'] = pxePassword
            kwargs['scp_hostname'] = pxeHostname
            kwargs['scp_dir_root'] = scpDirRoot
            kwargs['http_dir_root'] = httpDirRoot
            kwargs['http_dir_tmp'] = httpDirTmp
        else:
            kwargs['scp_username'] = pxe_username
            kwargs['scp_password'] = pxe_password
            kwargs['scp_hostname'] = pxe_hostname
            kwargs['scp_dir_root'] = kwargs.get('scp_dir_root',
                                                '{}/{}'.format(pxe_dir['scp_dir_root'], branch))
            kwargs['http_dir_root'] = kwargs.get('http_dir_root', pxe_dir['http_dir_root'])
            kwargs['http_dir_tmp'] = kwargs.get('http_dir_tmp', pxe_dir['http_dir_tmp'])
        kwargs['iso_image_path'] = iso_file_name
        kwargs['version_build'] = version
        self.series3fmc_baseline(**kwargs)

    def set_installation_timeouts(self, install_timeout=None, ssh_timeout=None):
        """Set the baseline timeout value for this device.

        :param install_timeout: in seconds, timeout for baseline procedure
        :param ssh_timeout: in seconds, timeout for ssh_vty connection
        :return: None
        """

        self.installation_timeout = install_timeout or MAX_TIME_INSTALL
        logger.info('setting baseline max timeout to {}'.format(self.installation_timeout))
        self.ssh_timeout = ssh_timeout or SSH_DEFAULT_TIMEOUT
        logger.info('setting ssh vty max timeout to {}'.format(self.ssh_timeout))

    def baseline_using_serial(self, iso_url, mgmt_ip, mgmt_netmask, mgmt_gateway, mgmt_intf="eth0",
                              power_cycle_flag=False, pdu_ip='', pdu_port='',
                              pdu_user='admn', pdu_pwd='admn', uut_ip6=None, uut_prefix=None,
                              uut_gateway6=None, timeout=None, change_password=True, ipv4_mode='static',
                              ipv6_mode='', dns_servers='', search_domains='', ntp_servers=''):
        """ Baseline Series3FMC device through its physical serial port connection.

        :param iso_url: http url of iso image
                http://10.83.65.25/cache/Development/6.2.3-1612/iso/Sourcefire_Defense_Center_M4-6.2.3-1612-Restore.iso
        :param mgmt_ip: management interface ip address
        :param mgmt_netmask: management interface netmask
        :param mgmt_intf: management interface gateway
        :param mgmt_gateway: management interface gateway
        :param timeout: in seconds; time to wait for device to boot with the new image;
                        if not provided, default baseline time is 7200s
        :param power_cycle_flag: True power cycle before baseline, False otherwise
        :param pdu_ip: string of IP addresses of the PDU's
        :param pdu_port: string of power port on the PDU's
        :param pdu_user: usernames for power bar servers
        :param pdu_pwd: passwords for power bar servers
        :param uut_ip6: Device IPv6 Address
        :param uut_prefix: Device IPv6 Prefix
        :param uut_gateway6: Device IPv6 Gateway
        :param timeout: in seconds; time to wait for device to boot with the new image;
                        if not provided, default baseline time is 7200s
        :param change_password: Flag to change the password after baseline
        :param ipv4_mode: IPv4 configuration mode - 'static' or 'dhcp'
        :param ipv6_mode: IPv6 configuration mode - 'dhcp', 'router' or 'manual'
        :param dns_servers: a comma-separated string of DNS servers
        :param search_domains: a comma-separated string of search domains
        :param ntp_servers: a comma-separated string of NTP servers
        :return:

        """

        publish_kick_metric('device.series3fmc.baseline', 1)
        # set baseline timeout
        self.set_installation_timeouts(timeout)

        # network configuration flag used to indicate if cli configuration is
        # still needed after baseline
        network_config = False

        self.iso_file = iso_url.split('/')[-1]
        self.uut_ip = mgmt_ip
        self.uut_netmask = mgmt_netmask
        self.uut_gateway = mgmt_gateway
        self.uut_ip6 = uut_ip6
        self.uut_prefix = uut_prefix
        self.uut_gateway6 = uut_gateway6

        if power_cycle_flag:
            self.power_cycle(pdu_ip, pdu_port, pdu_user, pdu_pwd)
            d0 = Dialog([
                ['boot:', None, None, False, False],
            ])
            d0.process(self.spawn_id, timeout=600)

            self._boot_selection(self.spawn_id, self.sm.patterns.prompt.lilo_boot_menu_prompt)
            self._move_from_lilo_boot_menu_to_lilo_os()
        else:
            self._try_to_goto_prompt(prompt='any')
            # Raise exception if device is not in below states
            baseline_valid_states = ["prelogin_state", "admin_state", "sudo_state", "lilos_state", "lilobootmenu_state",
                                     "fireos_state"]

            if self.sm.current_state not in baseline_valid_states:
                msg = "Baseline Serial function needs one of the valid state from %s" % baseline_valid_states
                logger.error(msg)
                raise RuntimeError(msg)

            # ---------------------------------------------
            # Bring the system to liloos_state
            # ---------------------------------------------
            if self.sm.current_state in ["lilobootmenu_state"]:
                self._move_from_lilo_boot_menu_to_lilo_os()

            # go to sudo_state
            elif self.sm.current_state in ["admin_state", "prelogin_state", "fireos_state"]:
                self.go_to('sudo_state')

            # reboot the device, select 3-System Restore Mode and enter lilo_boot state
            if self.sm.current_state is 'sudo_state':
                self.spawn_id.sendline("reboot")

                d0 = Dialog([
                    ['reboot: machine restart', None, None, True, False],
                    ['boot:', None, None, False, False],
                ])
                d0.process(self.spawn_id, timeout=600)

                self._boot_selection(self.spawn_id, self.sm.patterns.prompt.lilo_boot_menu_prompt)
                # activate console
                self._move_from_lilo_boot_menu_to_lilo_os()

        # Here means device is in LILO OS State
        self.sm.update_cur_state('lilos_state')

        # Try install seq with retry for 5 times
        for i in range(1, 5):
            ret = self._install_from_iso(iso_url=iso_url, uut_ip=mgmt_ip, uut_netmask=mgmt_netmask,
                                         uut_gateway=mgmt_gateway, mgmt_intf=mgmt_intf,
                                         timeout=self.installation_timeout)

            if ret == 0:
                network_config = self.configuration_wizard(ipv4_mode=ipv4_mode, ipv6_mode=ipv6_mode,
                                                           ipv4=mgmt_ip, ipv4_netmask=mgmt_netmask,
                                                           ipv4_gateway=mgmt_gateway, ipv6=uut_ip6,
                                                           ipv6_prefix=uut_prefix, ipv6_gateway=uut_gateway6,
                                                           dns_servers=dns_servers, search_domains=search_domains,
                                                           ntp_servers=ntp_servers)
                break

            # If the device moved back to lilo boot menu. Try to go to lilo os and try install again
            if ret == 1:
                self._move_from_lilo_boot_menu_to_lilo_os()

        # Raise error in all the attempts failed
        if i >= 5:
            msg = "Failed to install sereis3Fmc using serial for %s times" % i
            logger.error(msg)
            raise RuntimeError(msg)

        time.sleep(60)

        logger.info('=== Validate mysql process')
        self.validate_mysql_process()

        if not network_config:
            logger.info('=== Configure network')
            self.configure_network(self.uut_gateway, self.uut_ip, self.uut_netmask,
                                   uut_ip6, uut_prefix, uut_gateway6,
                                   change_password)
            logger.info('=== Fully installed.')

        logger.info('=== Validate version')
        iso_file_name = iso_url.split('/')[-1]
        self.validate_version(iso_file_name=iso_file_name)

        logger.info('Installation completed successfully.')

    def _boot_selection(self, spawn, prompt):

        spawn.send("\t")
        spawn.expect('boot:')
        spawn.send("\t")
        spawn.expect('boot:')
        spawn.send("Restore_Serial\r")
        spawn.expect(prompt, timeout=120)

    def baseline_by_branch_and_version_private_image(self, fmc, site, branch, version,
                                       uut_ip,uut_netmask, uut_gateway,uut_password,iso_file_name,private_image_server,
                                       dns_server, iso_file_type='Restore',
                                       timeout=None, **kwargs):
        """Baseline series3 fmc by branch and version using PXE servers on.
        private ISO

        Look for needed files on devit-engfs, copy them to the local kick server
        and use them to baseline the device.

        :param fmc: object of Series3fmc for reconnecting the device
        :param site: e.g. 'ful', 'ast', 'bgl'
        :param branch: branch name, e.g. 'Release', 'Feature'
        :param version: software build-version, e,g, 6.2.3-623
        :param uut_ip: Device IP Address
        :param uut_netmask: Device Netmask
        :param uut_gateway: Device Gateway
        :param iso_file_name:Private iso image e.g '/scratch/moasano/Cisco_Firepower_Mgmt_Center_Virtual-6.5.0-1128.iso'
        :param private_image_server:Private server on which the iso is placed. The same iso will be used to baseline the vFMC.
        :param dns_server: DNS server
        :param iso_file_type: 'Autotest' or 'Restore'; defaulted to 'Restore'
        :param timeout: in seconds; time to wait for device to boot with the new image;
                        if not provided, default baseline time is 7200s
        :param \**kwargs:
        :Keyword Arguments, any of below optional parameters:
        :param hostname: hostname to be set, defaulted to 'firepower'
        :param search_domains: search domains delimited by comma,
                                defaulted to 'cisco.com'
        :param detection_mode: defaulted to 'Inline',
                                can be: 'Passive','Inline','Access Control','Network Discovery'
        :param http_dir_root: absolute path of the http root
                                e.g. '/tftpboot/asa/'
        :param http_dir_tmp: the temp dir under http_dir_root for version_build
                            and the config file, e.g. '/cache/temp_config'
        :param scp_dir_root: absolute path of the dir for version_build
                            e.g. '/tftpboot/asa/cache/Release' or '/tftpboot/asa/cache/Development'
        :param bz_image: bz image has to be installed on DUT, which is copied
                        from the scp_server, defaulted to 'bzImage*'
        :param usb_image: usb image has to be installed on DUT, which is copied
                            from the scp_server, defaulted to 'usb-ramdisk*'
        :param arch: arch type, defaulted to 'x86_64'
        :param console: set to True if the device is accessed through its console;
                        defaulted to False
        :return: None
        """
        kwargs['fmc'] = fmc
        kwargs['uut_ip'] = uut_ip
        kwargs['uut_password'] = uut_password
        kwargs['uut_netmask'] = uut_netmask
        kwargs['uut_gateway'] = uut_gateway
        kwargs['timeout'] = timeout
        kwargs['dns_server'] = dns_server
        arch = kwargs.get('arch', 'x86_64')
        logger.info('=== Feaching bz and usb file from pse or engfs ...')
        server_ip, tftp_prefix, scp_prefix, files = prepare_installation_files(site, 's3fmc', branch, version,
                                                                               iso=iso_file_type, arch_type=arch)
        if not files[0]:
            raise Exception('Series3 FMC iso file not found on server')

        logger.debug('=== server {}, tftp prefix {}, scp prefix {}'.format(server_ip, tftp_prefix, scp_prefix))
        kwargs['http_server'] = private_image_server
        kwargs['scp_server'] = server_ip
        kwargs['scp_port'] = '22'
        kwargs['scp_username'] = pxe_username
        kwargs['scp_password'] = pxe_password
        kwargs['scp_hostname'] = pxe_hostname
        kwargs['scp_dir_root'] = kwargs.get('scp_dir_root',
                                                '{}/{}'.format(pxe_dir['scp_dir_root'], branch))
        kwargs['http_dir_root'] = kwargs.get('http_dir_root', pxe_dir['http_dir_root'])
        kwargs['http_dir_tmp'] = kwargs.get('http_dir_tmp', pxe_dir['http_dir_tmp'])
        logger.info('=== Private ISO provided by user ...')
        kwargs['iso_image_path'] = iso_file_name
        kwargs['version_build'] = version
        kwargs['uut_ip'] = uut_ip
        kwargs['dns_server'] = dns_server
        logger.info('=== Files collected, Start Baseline vFMC with private iso ...')
        self.series3fmc_baseline_private_image(**kwargs)

    def series3fmc_baseline_private_image(self, fmc, http_server, scp_server, scp_port,
                            scp_username, scp_password, scp_hostname,
                            version_build, iso_image_path,
                            uut_ip, uut_password,uut_netmask, uut_gateway, dns_server,
                            hostname='firepower',
                            search_domains='cisco.com',
                            http_dir_root='/var/www',
                            http_dir_tmp='iso/tmp_dev',
                            scp_dir_root='/nfs/netboot/ims/Release',
                            bz_image='bzImage*',
                            usb_image='usb-ramdisk*',
                            arch='x86_64',
                            console=False,
                            timeout=None, http_port='80'):
        """Install FMC iso image on Series 3 device via http.

        :param fmc: object of Series3fmc for reconnecting the device
        :param http_server: HTTP Server IP Address is private ISO server address, e.g. '10.83.68.41'
                            must not use the same http server as the scp server
        :param http_port: HTTP server port, e.g. '80'
        :param scp_server: SCP Server IP Address, e.g. '10.89.23.80'. This is PXE Server address
        :param scp_port: SCP Server Port, e.g. '22'
        :param scp_username: SCP Username
        :param scp_password: SCP Password
        :param scp_hostname: SCP Hostname
        :param version_build: e.g. '6.2.1-1366'
        :param iso_image_path: FMC image path on HTTP server/private HTTP server
                                e.g. '/scratch/moasano/Cisco_Firepower_Mgmt_Center_Virtual-6.5.0-1128.iso'
        :param uut_ip: Device IP Address
        :param uut_netmask: Device Netmask
        :param uut_gateway: Device Gateway
        :param dns_server: DNS server
        :param hostname: hostname to be set, e.g. 'BATIT-3D8130-1-AST'
        :param search_domains: search domains delimited by comma,
                                defaulted to 'cisco.com'
        :param http_dir_root: absolute path of the http root
                                e.g. '/var/www'
        :param http_dir_tmp: the temp dir under http_dir_root for version_build
                            and the config file, e.g. 'iso/tmp_dev'
        :param scp_dir_root: absolute path of the dir for version_build
                            e.g. '/nfs/netboot/ims/Release' or '/nfs/netboot/ims/Development'
        :param bz_image: bz image has to be installed on DUT, which is copied
                        from the scp_server, defaulted to 'bzImage*'
        :param usb_image: usb image has to be installed on DUT, which is copied
                            from the scp_server, defaulted to 'usb-ramdisk*'
        :param arch: arch type, defaulted to 'x86_64'
        :param console: set to True if the device is accessed through its console;
                        defaulted to False
        :param timeout: in seconds; time to wait for device to boot with the new image;
                        if not provided, default baseline time is 7200s
        :return: None

        """
        publish_kick_metric('device.series3fmc.baseline', 1)
        # set baseline timeout
        self.set_installation_timeouts(timeout)
        # Define attributes and variables
        self.fmc = fmc
        self.http_server = http_server
        self.http_port = http_port
        self.scp_server = scp_server
        self.scp_port = scp_port
        self.scp_username = scp_username
        self.scp_password = scp_password
        self.scp_hostname = scp_hostname
        self.version_build = version_build
        self.iso_image_path = iso_image_path
        self.iso_file = iso_image_path.split('/')[-1]
        self.uut_ip = uut_ip
        self.uut_password = uut_password
        self.uut_netmask = uut_netmask
        self.uut_gateway = uut_gateway
        self.dns_server = dns_server
        self.hostname = hostname
        self.search_domains = search_domains
        self.http_dir_root = http_dir_root
        self.http_dir_tmp = http_dir_tmp
        self.http_dir_for_build = '{}/{}/{}'.format(self.http_dir_root,
                                                    self.http_dir_tmp, self.version_build)
        self.config_file = '{}.config'.format(self.iso_file.strip('.iso'))
        self.config_path = '{}/{}'.format(self.http_dir_for_build, self.config_file)
        self.config_file_url = '{}/{}/{}'.format(self.http_dir_tmp, self.version_build, self.config_file)
        self.scp_dir_root = scp_dir_root
        self.bz_image = bz_image
        self.usb_image = usb_image
        self.arch = arch
        self.bz_image_path = '{}/{}/os/{}/boot/{}'.format(self.scp_dir_root,
                                                          self.version_build,
                                                          self.arch, self.bz_image)

        self.usb_image_path = '{}/{}/os/{}/ramdisks/{}'.format(self.scp_dir_root,
                                                               self.version_build,
                                                               self.arch, self.usb_image)

        self.console = console

        logger.info('=== Generate config file on the http server ...')
        self.generate_config_file(private_iso=True)
        logger.info('=== Copy bz and usb files to dut ...')
        logger.info('=== Generate lilo file on dut ...')
        self.go_to('sudo_state')
        self.copy_bz_usb_files_setup_lilo_to_install(private_iso=True)

        if not self.console:
            logger.info('=== Confirm device is rebooted ...')
            self.confirm_device_rebooted()
            logger.info('=== Wait for device to be up and configure device ...')
            self.poll_device_and_configure(initial_password=self.uut_password)
            if self.uut_password is not self.sm.patterns.default_password:
                conn = self.fmc.ssh_vty(ip=Series3Constants.uut_ssh_ip,
                                        port=Series3Constants.uut_ssh_port,
                                        password=self.uut_password,
                                        timeout=self.ssh_timeout)
            else:
                conn = self.fmc.ssh_vty(ip=Series3Constants.uut_ssh_ip,
                                        port=Series3Constants.uut_ssh_port,
                                        password=self.sm.patterns.default_password,
                                        timeout=self.ssh_timeout)
            self.spawn_id = conn.spawn_id
        else:
            logger.info('=== Console connection available, wait for login prompt ...')
            d = Dialog([
                ['login:', 'sendline({})'.format(self.sm.patterns.login_username),
                 None, True, False],
                ['Password:', 'sendline({})'.format(self.sm.patterns.default_password),
                 None, False, False],
            ])
            d.process(self.spawn_id, timeout=self.installation_timeout)
        logger.info('=== Fully installed.')
        logger.info('=== Validate version installed ...')
        self.validate_version(self.iso_file,private_iso=True)





import time

from unicon.eal.dialogs import Dialog


def clear_error(spawn):
   time.sleep(0.1)
   spawn.sendline('\x15')
   time.sleep(0.5)
   spawn.sendline('')
   spawn.sendline('')


CONFIGURATION_DIALOG = Dialog([
   ['Press <ENTER> to display the EULA: ', 'sendline()', None, True, False],
   ['--More--', 'send(\x20)', None, True, False],
   ["Please enter 'YES' or press <ENTER> to AGREE to the EULA: ", 'sendline()', None, True, False],
   [r'\x07$', clear_error, None, True, True],
   ['> ', 'sendline(expert)', None, True, False],
   ['[#$] ', None, None, False, False],
   ['(F|f)irepower-module\d+>', None, None, False, False],
   ['Would you like to configure this system', 'sendline(FMC)', None, True, False],
   ['This selection cannot be undone', 'sendline(y)', None, True, False],
   ['Enter a fully qualified hostname for this system', 'sendline()', None, True, False],
   ['Configure IPv4 via DHCP or static configuration', 'sendline()', None, True, False],
   ['Enter an IPv4 address for the management interface', 'sendline()', None, True, False],
   ['Enter an IPv4 netmask for the management interface', 'sendline()', None, True, False],
   ['Enter the IPv4 default gateway for the management interface', 'sendline()', None, True, False],
   ['Do you want to configure IPv6', 'sendline()', None, True, False],
   ['Configure IPv6 via DHCP, router, or manually', 'sendline()', None, True, False],
   ['Enter the IPv6 address for the management interface', 'sendline()', None, True, False],
   ['Enter the IPv6 address prefix for the management interface', 'sendline()', None, True, False],
   ['Enter the IPv6 gateway for the management interface', 'sendline()', None, True, False],
   ['Enter a comma-separated list of DNS servers', 'sendline()', None, True, False],
   ['Enter a comma-separated list of search domains', 'sendline()', None, True, False],
   ['Enter a comma-separated list of NTP servers', 'sendline()', None, True, False],
   ['Are these settings correct', 'sendline(y)', None, True, False],
   ['Updated network configuration', None, None, True, False],
   ['Successfully set NTP configuration', None, None, True, False]
])

TYPE_TO_STATE_MAP = {
    "WmLine": "fxos_state",
    "KpLine": "fxos_state",
    "SspLine": "mio_state",
    "ChassisLine": "mio_state"
}

# list of line classes which are qualified for getting the hostname dynamically at
# ssh connection
DEVICE_LIST = ["<class 'kick.device2.ftd5500x.actions.ftd5500x.Ftd5500xLine'>",
               "<class 'kick.device2.ssp.actions.ssp.SspLine'>",
               "<class 'kick.device2.fmc.actions.fmc.FmcLine'>" ]



"""Methods to check whether device is available via its console."""

import logging
import time

try:
    from kick.misc.convert_bytes import string_to_bytes
except ImportError:
    from kick.miscellaneous.convert import string_to_bytes

from unicon.eal.dialogs import Dialog
from kick.miscellaneous.credentials import *
from unicon.eal.expect import Spawn, TimeoutError

LOGGER = logging.getLogger(__name__)
HANDLER = logging.StreamHandler()
FORMATTER = logging.Formatter('%(asctime)s : %(levelname)s : %(name)s : %(message)s')
HANDLER.setFormatter(FORMATTER)
LOGGER.addHandler(HANDLER)

DEFAULT_TIMEOUT = 60
DEFAULT_USERNAME = 'myusername'
DEFAULT_PASSWORD = 'mypassword'
DEFAULT_ENPASSWORD = 'myenpassword'

VALID_PROMPTS = ['ciscoasa>', '\r\nciscoasa:~\$', '.*login:', 'rommon.*>', '.*-boot>', '\r\n[\x07]?>', '.*# $',
                 '\r\n.*@.*\$ $']

def wait_until_available(host, port, timeout=300, user=DEFAULT_USERNAME,
                         pwd=DEFAULT_PASSWORD, prompt='login:', access='telnet'):
    """Wait until device is available up to the timeout. If not, raise an
    exception.

    :param host: Ip of the device/console
    :param port: console port
    :param timeout: timeout
    :param user: username
    :param pwd: password
    :param prompt: expected prompt
    :param access: type of access: telnet or ssh
    :return: True if device is available, False if it's not

    """

    LOGGER.info('Wait until %s on port %s is available; timeout=%d'
                % (host, port, timeout))

    if user == DEFAULT_USERNAME:
        user = get_username(user)
    if pwd == DEFAULT_PASSWORD:
        pwd = get_password(pwd)

    first_time = True
    start_time = time.time()
    while first_time or (start_time + timeout > time.time()):
        if is_available(host, port, user, pwd, prompt, access):
            LOGGER.info('%s on port %s is available in %d seconds' %
                        (host, port, round(time.time() - start_time, 1)))
            return True
        else:
            time.sleep(5)
            first_time = False
    LOGGER.error('Cannot connect to %s on port %d in %d seconds' %
                 (host, port, timeout))


def is_available(host, port, user=DEFAULT_USERNAME, pwd=DEFAULT_PASSWORD,
                 prompt='firepower login:', access='telnet'):
    """Checks whether device is available.

    :param host: Ip of the device/console
    :param port: console port
    :param user: username
    :param pwd: password
    :param prompt: expected prompt
    :param access: type of access: telnet or ssh
    :return: True if device is available, False if it's not

    """

    if user == DEFAULT_USERNAME:
        user = get_username(user)
    if pwd == DEFAULT_PASSWORD:
        pwd = get_password(pwd)

    VALID_PROMPTS.append(prompt)
    if access == 'telnet':
        spawn_id = Spawn('telnet {} {}\n'.format(host, port))
        try:
            spawn_id.expect(
                "Connected to.*Escape character is '\^\]'\..*Username: ")
        except OSError:
            if port is not 23:
                clear_line(host=host, port=int(port) % 100, user=user, pwd=pwd, en_password=pwd)
            spawn_id = Spawn('telnet {} {}\n'.format(host, port))
            try:
                spawn_id.expect("Connected to.*Escape character is '\^\]'\..*Username: ")
            except OSError:
                spawn_id.close()
                return False
        spawn_id.sendline(user)
        spawn_id.expect("Password: ")
        spawn_id.sendline(pwd)
        try:
            spawn_id.expect("Password OK.*")
        except TimeoutError:
            LOGGER.debug("'Password OK' message did not appear ... continue")
        spawn_id.sendline('')
        try:
            __wait_for_rommon(spawn_id, 900)
        except:
            LOGGER.info("\nFailed to get a valid prompt")
            spawn_id.close()
            return False
        LOGGER.info('%s on port %d is available' % (host, port))
        spawn_id.close()
    elif access == 'ssh':
        try:
            if port is not 22:
                clear_line(host=host, port=port % 100, access='ssh', user=user, pwd=pwd, en_password=pwd)
            spawn_id = Spawn('ssh -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no -l {} -p {} {}'.
                             format(user, port, host))
            d1 = Dialog([
                ['continue connecting (yes/no)?', 'sendline(yes)', None, True, False],
                ['(P|p)assword:', 'sendline({})'.format(pwd), None, False, False],
                ['Connection refused', None, None, False, False],
            ])
            d1.process(spawn_id, timeout=60)
            try:
                spawn_id.expect("Password OK.*")
            except:
                pass
            spawn_id.sendline()
            time.sleep(10)
            try:
                __wait_for_rommon(spawn_id, 900)
            except:
                LOGGER.info("\nFailed to get a valid prompt")
                spawn_id.close()
                return False
            LOGGER.info('%s on port %d is available' % (host, port))
            spawn_id.close()
        except:
            return False
    else:
        raise RuntimeError('Device can be accessed only by telnet or ssh')

    return True


def __wait_for_rommon(spawn_id, timeout):
    # The system will reboot, wait for the following prompts
    d = Dialog([['Use BREAK or ESC to interrupt boot', 'sendline({})'.format(chr(27)), None, True, False],
                ])
    
    for prompt in VALID_PROMPTS:
        d.append([prompt, None, None, False, False], )

    d.process(spawn_id, timeout=timeout)


def tty_id_from_roty_id(spawn, port):
    """
        Function that determines the tty identifier from the roty port
        :param spawn: the spawn connection
        :param port: the roty port
        :return: the tty identifier for the line
    """
    spawn.sendline('terminal length 30\n')
    spawn.sendline('show line\n')
    time.sleep(3)
    output_lines = []
    # parse show line output into lines list
    while True:
        match_output = spawn.expect('.*').match_output
        output_lines.extend(match_output.split('\n'))
        if '--More--' in match_output:
            spawn.sendline('\x20')
            time.sleep(3)
        else:
            break

    # determine header column indexes
    header = None
    header_roty_index = None
    header_tty_index = None
    for i in range(len(output_lines)):
        # only determine the header once, the first
        # encountered header
        output_lines[i] = output_lines[i].replace('--More--', '')
        output_lines[i] = output_lines[i].replace('\b', '')
        output_lines[i] = output_lines[i].strip()
        if not header:
            if 'Tty' in output_lines[i]:
                header = tuple(output_lines[i].split())
                for j in range(len(header)):
                    if header[j] == 'Roty':
                        header_roty_index = j
                    if header[j] == 'Tty':
                        header_tty_index = j
        if len(output_lines[i]) > 0:
            if output_lines[i][0] == '*':
                # when the line starts with * and the *
                # is not separated by a space from the following
                # column we need to separate it manually
                # for consistency
                if output_lines[i][1] != ' ':
                    output_lines[i] = output_lines[i][0] + ' ' + output_lines[i][1:]
            else:
                # and make uniform the rest of the lines also with a dummy character
                # instead of *
                output_lines[i] = '@ ' + output_lines[i][:]

    # if the header is not present we can't reliably
    # determine the fields
    if not header:
        raise RuntimeError('no Tty column found in show line header.')

    # if the roty and tty fields are not displayed
    # we can't realiably determine the line
    if header_roty_index is None or header_tty_index is None:
        raise RuntimeError('show line is not showing the Tty and Roty columns.')

    # offsetting the columns accordingly to account
    # for * and @ at the start of the lines
    header_roty_index += 1
    header_tty_index += 1

    # determine tty line id from roty id
    line_id = None
    for line in output_lines:
        fields = tuple(line.split())
        try:
            if int(fields[header_roty_index]) == port:
                line_id = fields[header_tty_index]
                break
        except:
            # if we cannot convert the roty id to an int
            # we consider that things are not correctly configured
            # and go to the next output line and continue searching
            pass
    return line_id


def clear_line(host, port, user=DEFAULT_USERNAME, pwd=DEFAULT_PASSWORD, prompt='#',
               access='telnet', en_password=DEFAULT_ENPASSWORD, timeout=None):
    """Clear line corresponding to a device; this is required because only a
    single console connection is available.

    If somebody or some process failed to close the connection, it
    should be cleared explicitly.
    
    This function accepts only ssh and telnet connections.

    :param host: ip address of terminal server
    :param port: device line number in terminal server to be cleared
                for example, if port 2005 is mapped to line 5, port=5
    :param user: username
    :param pwd: password
    :param prompt: expected prompt after logging in
    :param access: ssh or telnet; default is set to telnet
    :param en_password: enable password to switch to line configuration mode
    :param timeout: how long the connection and authentication would take in seconds;
                    if not provided, default is 60s
    :return: None
    
    """

    if user == DEFAULT_USERNAME:
        user = get_username(user)
    if pwd == DEFAULT_PASSWORD:
        pwd = get_password(pwd)
    if en_password == DEFAULT_ENPASSWORD:
        en_password = get_password(en_password)

    if not timeout:
        timeout = DEFAULT_TIMEOUT

    d1 = None
    spawn = None

    # establish a connection to the terminal server
    if access == 'telnet':
        spawn = Spawn('telnet {} {}'.format(host, '23'))
        d1 = Dialog([
            ["Connected to.*Escape character is '\^\]'\.", 'sendline()', None, True, False],
            ['.*Username:', 'sendline({})'.format(user), None, True, False],
            ['(p|P)assword:', 'sendline({})'.format(pwd), None, True, True],
            [prompt, 'sendline()', None, False, False],
            ['>', 'sendline()', None, False, False],
        ])

    elif access == 'ssh':
        spawn = Spawn('ssh -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no '
                      '-l {} -p {} {}'.format(user, '22', host))
        d1 = Dialog([
            ['continue connecting (yes/no)?', 'sendline(yes)', None, True, False],
            ['(p|P)assword:', 'sendline({})'.format(pwd), None, False, False],
        ])

    else:
        LOGGER.error('Unknown protocol: Telnet or ssh supported only')

    try:
        LOGGER.info('Trying to connect to {}'.format(host))
        d1.process(spawn, timeout=timeout)
        try:
            spawn.expect("Password OK.*")
        except TimeoutError:
            LOGGER.info("'Password OK' message didn't appear")
            pass
        spawn.sendline()
    except TimeoutError:
        LOGGER.error('Failed to connect to terminal server')
        raise Exception('Failed to connect to terminal server')

    # clear port section
    try:
        spawn.expect('#')
    except:
        # expect >
        spawn.sendline('en')
        try:
            spawn.expect('Password:')
            spawn.sendline(en_password)
        except:
            pass
    try:
        line_id = tty_id_from_roty_id(spawn, port)
        LOGGER.info('detected line number for clearing: {} from port {}'.
                    format(line_id, port))
        if line_id:
            spawn.sendline('clear line {}'.format(line_id))
            spawn.expect('[confirm]')
            spawn.sendline('')
            spawn.expect('[OK]')
            LOGGER.info('line: {} was cleared'.format(port))
        spawn.close()
    except TimeoutError:
        spawn.close()
        LOGGER.error('Line: {} was not cleared'.format(port))
        raise Exception('Line {} was NOT cleared'.format(port))



import munch


class Elektra5585Patterns:
    def __init__(self, hostname, login_password, sudo_password, enable_password):
        self.hostname = hostname
        self.username = 'admin'
        self.login_password = login_password
        self.sudo_password = sudo_password
        self.enable_password = enable_password
        self.default_password = 'Admin123'

        self.prompt = munch.Munch()

        self.prompt.password_prompt = '\r\n[Pp]assword: '
        self.prompt.prelogin_prompt = '\r?\n?({}|ciscoasa) login: '.format(self.hostname)
        self.prompt.pkglogin_prompt = '\r?\n?(SSP[1246]0SF|firepower|asasfr) login: '
        self.prompt.fireos_prompt = '\r\n[\x07]?> '
        self.prompt.system_prompt = '\r\n[\x07]?system> '
        self.prompt.expert_prompt = '\r\n[^\r\n]*admin@.*?\$ '
        self.prompt.sudo_prompt = '\r\n[^\r\n]*(root@.*#) '
        self.prompt.rommon_prompt = 'rommon.*>'
        self.prompt.firepower_boot = '-boot>'
        self.prompt.config_prompt = '[\r\n]({}|firepower)[\w]*\([\w\-]+\)# '.format(self.hostname)



# python kick/device2/fxos/tests/baseline_ssp.py --testbed kick/device2/fxos/testbed_samples/
# ssp3ru_native.yaml --build 1119 --version 6.5.0 --branch Development --fxos_file fxos-k9.2.7.1.4.SPA

import logging
import sys
import traceback

from ats import aetest
from ats import topology

from kick.device2.fxos.app import AppInstance
from kick.device2.fxos.util_app import AppArgumentsParser
from kick.device2.ssp.actions import Ssp

KICK_EXTERNAL = False

try:
    from kick.file_servers.file_servers import *
except ImportError:
    KICK_EXTERNAL = True
    pass

from kick.device2.fxos.configuration_utils import ConfigurationClean, PackagePrerequisites, \
    ConfigureInterfacesUtils, ConfigureMultiInstanceUtils, InterfaceCleanUtils
from kick.device2.fxos.utils import create_dict

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

logging.getLogger('unicon').setLevel(logging.DEBUG)


def get_chassis_mio_console():
    global mio, MIO_CONSOLE
    if not MIO_CONSOLE:
        MIO_CONSOLE = mio.ssh_console(
            ip=SSH_CONSOLE_IP, port=SSH_CONSOLE_PORT, username=SSH_CONSOLE_USERNAME, password=SSH_CONSOLE_PASSWORD,
            timeout=120)
        # disable logout from the chassis
        MIO_CONSOLE.set_default_auth_timeouts()
    return MIO_CONSOLE


def get_args():
    return args


def get_build_no():
    return BUILD_NO


class CommonSetup(aetest.CommonSetup):
    groups = ['common', 'CommonSetup']

    def get_image_path(self, branch, version, build):
        BRANCHES = {
            #    Branch:   Top folder where builds for this branch are stored
            'RELEASE': 'Release',
            'TESTING': 'Testing',
            'DEVELOPMENT': 'Development',
            'MAIN': 'Development',
            'INTEG': 'Development',
            'INTEGRATION': 'Development',
            'DEVELOPMENT_SIT': 'Development',
            'DEVELOPMENT_ST': 'Development',
            'IMS_6_5_0': 'Development',
            'IMS_6_6_0': 'Development',
            'IMS_6_7_0': 'Development',
            'IMS_6_8_0': 'Development',
        }

        final_branch = branch
        final_build = build
        final_version = version
        feature_branch = False

        if final_branch.upper() not in BRANCHES.keys():
            log.debug("Branch '{}' is a Feature branch.".format(final_branch))
            feature_branch = True
            branch_value = 'Feature/{}'.format(branch)
        else:
            branch_value = BRANCHES[branch.upper()]

        version_build_text = "{}-{}{}".format(final_version, final_build, "." + final_branch if feature_branch else "")
        log.info("Using {} branch {}. Build folder: {}/{}.".format("feature " if feature_branch else "", final_branch,
                                                                   branch_value, version_build_text))

        return branch_value, version_build_text

    @aetest.subsection
    def main(self):
        import argparse

        parser = argparse.ArgumentParser(description="Baseline for QP devices.")

        parser.add_argument('--testbed', dest='testbed', required=True, help='path to the testbed file')
        parser.add_argument('--build', dest='build', required=True,
                            help='the build number; e.g. 1234, 10200')
        parser.add_argument('--version', dest='version', required=True, help='the version number; usage e.g. 6.2.3')
        parser.add_argument('--branch', dest='branch', required=True,
                            help='the branch name; only feature branches are case sensitive (e.g. GL_MCTX_PHASE0, SNORT3); the other branches are case insensitive')
        parser.add_argument('--report_build', dest='report_build',
                            help='the build number for the report; if missing it will take the value from --build; usage e.g. 1234, 10200')
        parser.add_argument('--report_version', dest='report_version',
                            help='the version number for the report; if missing it will take the value from --version; usage e.g. 6.2.3')
        parser.add_argument('--report_branch', dest='report_branch',
                            help='the branch name for the report; if missing it will take the value from --branch; feature branches are case sensitive; DEVELOPMENT, MAIN are not case sensitive; the value you provide here will be seen on the Kick UI in the reports section')
        parser.add_argument('--report_type', dest='report_type',
                            help='if specified it will update the result of the test in Kick reporting; e.g. BQT/BAT_2_0 (if you need a new report type (that doesn\'t yet exist) you need to contact Kick team)')
        parser.add_argument('--report_subtype', dest='report_subtype',
                            help='if specified it will update the result of the test in Kick reporting; e.g. BQT or BAT')
        parser.add_argument('--report_env_type', dest='report_env_type',
                            help='if specified it will update the result of the test in Kick reporting; e.g. FMC_QW_1150')
        parser.add_argument('--report_group_by', dest='report_group_by',
                            help='this is needed when you want to report multiple execution plans in the same cell')
        parser.add_argument('--fxos_file', dest='fxos_file', required=False, help='the fxos image to be used')
        parser.add_argument('--apps', dest='apps', required=False,
                            help='if specified it will create only the apps in the list; usage e.g. 1-3,6-9 or ftdapp7,ftdapp25')
        parser.add_argument('--cpu_core_count', dest='cpu_core_count', required=False,
                            help='if specified it will override the resource profile we have in testbed with new resource profiles;'
                                 'usage e.g. --cpu_core_count 6,8,12|6|6,10 or --cpu_core_count 6 or --cpu_core_count 6|12|6 '
                                 'each slot is separated by pipe |'
                                 'if we specify only one number it will be used for all the apps in the slot'
                                 'if we specify multiple numbers they will be used for the apps in order')

        global args
        args, unknown = parser.parse_known_args()
        if len(unknown) > 0:
            log.warning('Unknown arguments: {}'.format(unknown))

        global FXOS_FILE
        FXOS_FILE = None
        if args.fxos_file is not None:
            FXOS_FILE = args.fxos_file
        testbed_yaml = args.testbed
        tb = topology.loader.load(testbed_yaml)
        global SITE
        SITE = tb.custom.site
        global SSP_DEV
        SSP_DEV = tb.custom.device_name

        global BRANCH_PATH, VERSION_BUILD, VERSION, BUILD
        VERSION = args.version
        BUILD = args.build

        (BRANCH_PATH, VERSION_BUILD) = self.get_image_path(branch=args.branch, version=args.version, build=args.build)
        log.info("Using branch_path='{}' and version_build='{}'".format(BRANCH_PATH, VERSION_BUILD))

        global FTD_VERSION, BUILD_NO
        if re.match('^Feature\/', BRANCH_PATH):
            small_version_build = re.match('(.*\-.*)\.', VERSION_BUILD)
            FTD_VERSION = small_version_build.groups()[0]
            FTD_VERSION = re.sub('-', '.', FTD_VERSION)
            small_build = re.match('.*\-(.*)\.', VERSION_BUILD)
            BUILD_NO = small_build.groups()[0]
        else:
            FTD_VERSION = VERSION_BUILD
            FTD_VERSION = re.sub('-', '.', FTD_VERSION)
            BUILD_NO = VERSION_BUILD.split("-")[1]

        log.info('FTD_VERSION={}'.format(FTD_VERSION))
        log.info('BUILD_NO={}'.format(BUILD_NO))

        global HOSTNAME, SSH_CONSOLE_IP, SSH_CONSOLE_PORT, SSH_CONSOLE_USERNAME, SSH_CONSOLE_PASSWORD, SSH_CHASSIS_IP, \
            SSH_CHASSIS_PORT, SSH_CHASSIS_PORT, SSH_CHASSIS_USERNAME, SSH_CHASSIS_PASSWORD, NTP_SERVER, DNS_SERVERS, \
            SEARCH_DOMAINS, POWER_BAR_SERVER, POWER_BAR_PORT, CHASSIS_MGMT_IPV4_IP, CHASSIS_MGMT_IPV4_NETMASK, \
            CHASSIS_MGMT_IPV4_GATEWAY, CHASSIS_SOFTWARE, CHASSIS_INTERFACES, APPS, MIO_CONSOLE
        MIO_CONSOLE = None
        HOSTNAME = tb.devices[SSP_DEV].custom.hostname

        # although console is more intuitive, alt is still commonly used for console connection definition
        if hasattr(tb.devices[SSP_DEV].connections, 'alt'):
            SSH_CONSOLE_IP = str(tb.devices[SSP_DEV].connections.alt.ip)
            SSH_CONSOLE_PORT = str(tb.devices[SSP_DEV].connections.alt.port)
            if hasattr(tb.devices[SSP_DEV].connections.alt, 'user'):
                SSH_CONSOLE_USERNAME = str(tb.devices[SSP_DEV].connections.alt.user)
            elif hasattr(tb.devices[SSP_DEV].connections.alt, 'username'):
                SSH_CONSOLE_USERNAME = str(tb.devices[SSP_DEV].connections.alt.username)
            SSH_CONSOLE_PASSWORD = str(tb.devices[SSP_DEV].connections.alt.password)
        elif hasattr(tb.devices[SSP_DEV].connections, 'console'):
            SSH_CONSOLE_IP = str(tb.devices[SSP_DEV].connections.console.ip)
            SSH_CONSOLE_PORT = str(tb.devices[SSP_DEV].connections.console.port)
            if hasattr(tb.devices[SSP_DEV].connections.console, 'user'):
                SSH_CONSOLE_USERNAME = str(tb.devices[SSP_DEV].connections.console.user)
            elif hasattr(tb.devices[SSP_DEV].connections.console, 'username'):
                SSH_CONSOLE_USERNAME = str(tb.devices[SSP_DEV].connections.console.username)
            SSH_CONSOLE_PASSWORD = str(tb.devices[SSP_DEV].connections.console.password)

        # CHASSIS MGMT DETAILS
        SSH_CHASSIS_IP = str(tb.devices[SSP_DEV].connections.management.ip)
        SSH_CHASSIS_PORT = str(tb.devices[SSP_DEV].connections.management.port)
        SSH_CHASSIS_USERNAME = str(tb.devices[SSP_DEV].connections.management.user)
        SSH_CHASSIS_PASSWORD = str(tb.devices[SSP_DEV].connections.management.password)

        try:
            NTP_SERVER = tb.devices[SSP_DEV].custom.ntp_server
        except:
            NTP_SERVER = None
        try:
            DNS_SERVERS = tb.devices[SSP_DEV].custom.chassis_network.get("dns_servers")
        except:
            DNS_SERVERS = None
        try:
            SEARCH_DOMAINS = tb.devices[SSP_DEV].custom.chassis_network.get("search_domains")
        except:
            SEARCH_DOMAINS = None

        POWER_BAR_SERVER = tb.devices[SSP_DEV].custom.chassis_power.get("power_bar_server")
        POWER_BAR_PORT = tb.devices[SSP_DEV].custom.chassis_power.get("power_bar_port")

        CHASSIS_MGMT_IPV4_IP = str(tb.devices[SSP_DEV].interfaces.chassis_mgmt.ipv4.ip)
        CHASSIS_MGMT_IPV4_NETMASK = str(tb.devices[SSP_DEV].interfaces.chassis_mgmt.ipv4.netmask)
        CHASSIS_MGMT_IPV4_GATEWAY = str(tb.devices[SSP_DEV].interfaces.chassis_mgmt.ipv4_gateway)

        CHASSIS_SOFTWARE = tb.devices[SSP_DEV].custom.chassis_software
        # need to insert in TB the startup_version from CLI
        instances = CHASSIS_SOFTWARE['applications']
        for i in instances:
            CHASSIS_SOFTWARE['applications'][i]["startup_version"] = FTD_VERSION
        CHASSIS_INTERFACES = tb.devices[SSP_DEV].interfaces
        del CHASSIS_INTERFACES["chassis_mgmt"]

        APPS = []
        app_parser = AppArgumentsParser()

        if args.apps is not None:
            APPS = app_parser.get_apps_list(args.apps)
        else:
            for app in CHASSIS_SOFTWARE['applications']:
                APPS.append(CHASSIS_SOFTWARE['applications'][app].get("application_identifier"))
        log.info("Will create only these apps: {}".format(APPS))

        log.info("Modifying the chassis software to take default values")
        CHASSIS_SOFTWARE = app_parser.build_app_template(CHASSIS_SOFTWARE, APPS)

        if args.cpu_core_count is not None:
            log.info("--cpu_core_count is specified so we need to override the resource profile")
            resource_profiles = {}
            count_per_slot = {}
            i = 1
            for slot_cpu_info in args.cpu_core_count.split("|"):
                app_cpu_info = slot_cpu_info.split(",")
                for app_cpu in app_cpu_info:
                    if app_cpu != "0" and app_cpu != "":
                        resource_profiles[app_cpu] = {}
                        resource_profiles[app_cpu]['name'] = 'stauto_{}'.format(app_cpu)
                        resource_profiles[app_cpu]['cpu_core_count'] = app_cpu
                    else:
                        log.info("Cannot use {} for resource profile! Skipping...".format(app_cpu))
                count_per_slot[i] = app_cpu_info
                i = i + 1
            CHASSIS_SOFTWARE['resource_profiles'] = resource_profiles
            for app_id in APPS:
                for i in CHASSIS_SOFTWARE['applications']:
                    if CHASSIS_SOFTWARE['applications'][i]['application_identifier'] == app_id:
                        slot = CHASSIS_SOFTWARE['applications'][i]['slot']
                        if len(count_per_slot[slot]) > 1:
                            CHASSIS_SOFTWARE['applications'][i]['resource_profile'] = 'stauto_{}'.format(
                                count_per_slot[slot].pop(0))
                        else:
                            CHASSIS_SOFTWARE['applications'][i]['resource_profile'] = 'stauto_{}'.format(
                                count_per_slot[slot][0])
                        break

    @aetest.subsection
    def create_ssp_device(self):
        global mio
        mio = Ssp(hostname=HOSTNAME,
                  login_username=SSH_CHASSIS_USERNAME,
                  login_password=SSH_CHASSIS_PASSWORD,
                  sudo_password=SSH_CHASSIS_PASSWORD,
                  power_bar_server=POWER_BAR_SERVER,
                  power_bar_port=POWER_BAR_PORT,
                  deploy_type='container',
                  app_identifier='dummy_name')

    @aetest.subsection
    def connect_to_chassis_mio_ssh(self):
        global dev
        dev = mio.ssh_vty(ip=SSH_CHASSIS_IP, port=SSH_CHASSIS_PORT, username=SSH_CHASSIS_USERNAME,
                          password=SSH_CHASSIS_PASSWORD, timeout=300)
        dev.reconnect_feature = {'enabled': True, 'max_retries': 3}


class CleanInstances(ConfigurationClean):
    groups = ['CleanMultiInstance']

    @aetest.test
    def init_params(self):
        self.handle = dev
        self.configuration = CHASSIS_SOFTWARE


class CleanInterfaces(InterfaceCleanUtils):
    groups = ['CleanInterfaces', 'chassis']

    @aetest.test
    def init_params(self):
        self.handle = dev


class PreparePrerequisites(PackagePrerequisites):
    groups = ['PreparePrerequisites']

    @aetest.test
    def init_params(self):
        self.handle = dev
        self.branch_path = BRANCH_PATH
        self.version = VERSION
        self.build = BUILD
        self.site = SITE
        self.fxos_file = FXOS_FILE
        self.version_build = VERSION_BUILD


class BaselineFXOS(aetest.Testcase):
    groups = ['BaselineFXOS', 'chassis']
    chassis_line = None

    @aetest.test
    def init_params(self):
        self.fxos_file = FXOS_FILE

    @aetest.test
    def upgrade_fxos(self):
        self.chassis_line = get_chassis_mio_console()
        if self.fxos_file:
            try:
                self.chassis_line.upgrade_bundle_package(self.fxos_file)
            except Exception as e:
                self.failed(
                    "Unable to install FXOS e= {} {} {}".format(e, traceback.format_exc(), sys.exc_info()),
                    goto=["common_cleanup"])


class ConfigureChassis(aetest.Testcase):
    groups = ['ConfigureChassis', 'chassis']
    chassis_line = None

    @aetest.test
    def configure_chassis(self, steps):
        self.chassis_line = get_chassis_mio_console()
        try:
            with steps.start("Change Chassis IP"):
                self.chassis_line.execute('top')
                self.chassis_line.execute('scope fabric a')
                self.chassis_line.execute('show detail')
                self.chassis_line.execute(
                    'set out-of-band ip {} netmask  {} gw {}'.format(CHASSIS_MGMT_IPV4_IP, CHASSIS_MGMT_IPV4_NETMASK,
                                                                     CHASSIS_MGMT_IPV4_GATEWAY))
                self.chassis_line.execute('commit-buffer')
                output = self.chassis_line.execute('show detail')
                if "IP Addr: ".format(CHASSIS_MGMT_IPV4_IP) not in output:
                    raise Exception("IPv4 was not set correctly")
                self.chassis_line.execute('up')
            if DNS_SERVERS is not None:
                with steps.start("Change DNS server"):
                    self.chassis_line.execute('scope system')
                    self.chassis_line.execute('scope services')
                    self.chassis_line.execute('create dns {}'.format(DNS_SERVERS))
                    self.chassis_line.execute('commit-buffer')
                    output = self.chassis_line.execute('show dns')
                    if "{}".format(DNS_SERVERS) not in output:
                        raise Exception("DNS Server was not set correctly")
            if SEARCH_DOMAINS is not None:
                with steps.start("Change domain name"):
                    self.chassis_line.execute('set domain-name {}'.format(SEARCH_DOMAINS))
                    self.chassis_line.execute('commit-buffer')
                    output = self.chassis_line.execute('show domain-name')
                    if "{}".format(SEARCH_DOMAINS) not in output:
                        raise Exception("Search Domain was not set correctly")
        except Exception as e:
            self.failed(
                "Unable to install FXOS e= {} {} {}".format(e, traceback.format_exc(), sys.exc_info()),
                goto=["common_cleanup"])

    @aetest.test
    def configure_ntp_server(self, steps):
        if NTP_SERVER:
            with steps.start("Change NTP server - if necessary"):
                self.chassis_line.execute('top')
                self.chassis_line.execute('scope system')
                self.chassis_line.execute('scope services')
                output = create_dict(self.chassis_line.execute('show ntp-server detail'))
                if not hasattr(output, 'Name') or output['Name'] != NTP_SERVER:
                    log.info("Changing NTP server")
                    self.chassis_line.execute('create ntp-server {}'.format(NTP_SERVER))
                    output = self.chassis_line.execute('commit-buffer')
                    if 'Error' in output:
                        raise Exception("Failed to change NTP server")
                    self.chassis_line.execute('up')
                    self.chassis_line.execute('show ntp-server detail')


class ConfigureInterfaces(ConfigureInterfacesUtils):
    groups = ['ConfigureInterfaces', 'chassis']

    @aetest.test
    def init_params(self):
        self.handle = dev
        self.interfaces = CHASSIS_INTERFACES


class ConfigureInstances(ConfigureMultiInstanceUtils):
    groups = ['ConfigureMultiInstance']

    @aetest.test
    def init_params(self):
        self.handle = dev
        self.configuration = CHASSIS_SOFTWARE
        self.apps_to_create = APPS


class InitializeApps(aetest.Testcase):
    groups = ['InitializeApps']

    def initialize_app(self, ssp_app):
        app_line = ssp_app.ssh_vty(ip=SSH_CHASSIS_IP,
                                   port=SSH_CHASSIS_PORT,
                                   username=SSH_CHASSIS_USERNAME,
                                   password=SSH_CHASSIS_PASSWORD,
                                   timeout=300)
        app_line.go_to("fireos_state")

    @aetest.test
    def init_apps(self):
        retry_app = []
        instances = CHASSIS_SOFTWARE['applications']
        for i in instances:
            if instances[i].get("application_identifier") in APPS:
                slot_id = instances[i].get("slot")
                deploy_type = instances[i].get("deploy_type")
                ssp_app = Ssp(hostname=HOSTNAME,
                              login_username=SSH_CHASSIS_USERNAME,
                              login_password=SSH_CHASSIS_PASSWORD,
                              sudo_password=SSH_CHASSIS_PASSWORD,
                              power_bar_server=POWER_BAR_SERVER,
                              power_bar_port=POWER_BAR_PORT,
                              slot_id=slot_id,
                              deploy_type=deploy_type,
                              app_identifier=instances[i].get("application_identifier"))
                try:
                    self.initialize_app(ssp_app)
                except:
                    log.error("Failed to initialize app. Checking if the blade is missing.")
                    app_instance = AppInstance(handle=dev, app_name=instances[i].get("application_name"),
                                               slot=instances[i].get("slot"),
                                               identifier=instances[i].get("application_identifier"))
                    app_details = app_instance.get_app_instance_detail()
                    if app_details['Oper State'] == "Not Available":
                        log.error("The blade is missing on slot {}. Skipping this app.".format(slot_id))
                    else:
                        retry_app.append(ssp_app)

        if len(retry_app) > 0:
            sleep_time = 60
            log.info("Sleeping {} seconds before retrying".format(sleep_time))
            time.sleep(sleep_time)
            for app in retry_app:
                try:
                    self.initialize_app(app)
                except Exception as e:
                    self.failed(
                        "Unable to install app for second retry e= {} {} {}".format(e, traceback.format_exc(),
                                                                                    sys.exc_info()),
                        goto=["common_cleanup"])


class CheckTCAMEntryLimit(aetest.Testcase):
    groups = ['CheckTCAMEntryLimit']

    @aetest.test
    def check_TCAM(self, steps):
        try:
            with steps.start("Get fabric-interconnect output"):
                dev.execute('top')
                dev.execute('scope chassis')
                dev.execute('scope fabric-interconnect')
                output = dev.execute('show detail')
                output = output.replace("\r\n", "")
            with steps.start("Check 'Ingress VLAN Group Entry Count'"):
                match = re.match(".+Ingress VLAN Group Entry Count \(Current\/Max\)\: (\d+)\/(\d+)", output)
                log.info("Ingress VLAN Group Entry Count Current = {}".format(match.group(1)))
                log.info("Ingress VLAN Group Entry Count Max = {}".format(match.group(2)))
                if int(match.group(1)) > int(match.group(2)):
                    raise Exception("Error: Ingress VLAN Group Entry Count - current greater than max")
            with steps.start("Check 'Switch Forwarding Path Entry Count'"):
                match = re.match(".+Switch Forwarding Path Entry Count \(Current\/Max\)\: (\d+)\/(\d+)", output)
                log.info("Switch Forwarding Path Entry Count Current = {}".format(match.group(1)))
                log.info("Switch Forwarding Path Entry Count Maxim = {}".format(match.group(2)))
                if int(match.group(1)) > int(match.group(2)):
                    raise Exception("Error: Switch Forwarding Path Entry Count - current greater than max")
        except Exception as e:
            self.failed(
                "Failed while getting TCAM info e= {} {} {}".format(e, traceback.format_exc(), sys.exc_info()),
                goto=["common_cleanup"])


class CommonCleanup(aetest.CommonCleanup):
    groups = ['common', 'CommonCleanup']

    @aetest.subsection
    def cleanup(self):
        dev.disconnect()


if __name__ == '__main__':
    aetest.main()



import re
import os
from urllib.parse import urljoin
import os.path
import requests
from bs4 import BeautifulSoup
from functools import reduce

import logging

logger = logging.getLogger(__name__)


def search_engfs_with_regex(site, branch, version, subdir='installers',
                            pattern='.*'):
    """
    find all files in a list that resides on the site engfs server, under
    branch/version/subdir, and matching pattern.

    :param site: such as 'ful' or a custom server, such as http://10.106.134.200/
    :param branch: such as 'Feature/SSL_OFFLOAD'
    :param version: such as '6.2.3-430.SSL_OFFLOAD'
    :param subdir: such as 'installers'
    :param pattern: this is the pattern to use in regex.
    :return: list of files matching the search
    """

    # url below will look like: https://firepower-engfs-sjc.cisco.com/\
    # netboot/ims/Development/6.3.0-80013/installers/, it doesn't have the
    # last part for file name, since we passed in '' as file_name.
    if site.startswith("http"):
        url = urljoin(site, os.path.join(os.path.join(branch, version, subdir)))
    else:
        url = _construct_devit_url(site, os.path.join(branch, version, subdir), '')
    logger.debug("searching in {}".format(url))

    # get content from url
    r = requests.get(url)
    data = r.text
    soup = BeautifulSoup(data, 'lxml')

    # find all valid images: we rely on the assumption that for a valid image,
    # its text content is the same string as its href, as well as its title
    images = []

    for link in soup.find_all('a'):
        if link.getText() == link.get('title') == link.get('href'):
            images.append(link.getText())

    logger.debug("list of all images for branch {} version {} subdir {}: {}"
                 "".format(branch, version, subdir, images))

    result = []
    p = re.compile(pattern)
    for image in images:
        if p.search(image):
            result.append(image)

    logger.debug("found matching images: {}".format(result))

    return result


def flattenlist(alist):
    """Remove empty lists/values and flatten the list

    :param alist: list or list of lists
    :return: flatten list
    """

    alist = filter(None, alist)
    return reduce(lambda x, y: x + y, alist, [])


def get_devices_dict(version, image=None, arch=None, feature=None):
    """Based on version and image, returns a dictionary containing
       the folder location and the patterns of the installation files for
       each device type

    :param version: build version, e.g. 6.2.3-623
    :param image: optional, 'Autotest' or 'Restore', required for M3, M4, S3 (FMC and Sensor)
    :param arch: optional, device architecture, required for S3 (FMC and Sensor) - e.g x86_64
    :param feature: optional, whether the build is on a feature branch (e.g. MARIADB)
    :return: a dictionary
    """

    if feature is None:
        feature = ''
    else:
        feature = ".{}".format(feature)

    devices = {
        'kenton': {'patterns': ['ftd-[\d\.-]+{}.pkg'.format(feature), 'ftd-boot-[\d.]+lfbff'],
                   'subdir': ['installers', 'installers/doNotRelease'],
                   },
        'saleen': {'patterns': ['ftd-[\d\.-]+{}.pkg'.format(feature), 'ftd-boot-[\d.]+cdisk'],
                   'subdir': ['installers', 'installers/doNotRelease'],
                   },
        'elektra': {'patterns': ['asasfr-sys-[\d.-]+.pkg', 'asasfr-5500x-boot-[\d.-]+img'],
                    'subdir': ['installers', 'installers/doNotRelease'],
                    },
        'm3': {'patterns': ['Sourcefire_Defense_Center_S3-{}{}-{}.iso'.format(version, feature, image),
                            'Sourcefire_Defense_Center-{}{}-{}.iso'.format(version, feature, image),
                            'Cisco_Firepower_Mgmt_Center-{}{}-{}.iso'.format(version, feature, image)],
               'subdir': ['iso', 'iso/doNotRelease'],
               },
        'm4': {'patterns': ['Sourcefire_Defense_Center_M4-{}{}-{}.iso'.format(version, feature, image),
                            'Cisco_Firepower_Mgmt_Center-{}{}-{}.iso'.format(version, feature, image),
                            'Sourcefire_Defense_Center-{}{}-{}.iso'.format(version, feature, image)],
               'subdir': ['iso', 'iso/doNotRelease'],
               },
        'm5': {'patterns': ['Sourcefire_Defense_Center-{}{}-{}.iso'.format(version, feature, image),
                            'Cisco_Firepower_Mgmt_Center-{}{}-{}.iso'.format(version, feature, image)],
               'subdir': ['iso', 'iso/doNotRelease'],
               },
        's3fmc': {'patterns': ['Sourcefire_Defense_Center_S3-{}{}-{}.iso'.format(version, feature, image),
                               'Cisco_Firepower_Mgmt_Center-{}{}-{}.iso'.format(version, feature, image)],
                  'subdir': ['iso', 'iso/doNotRelease'],
                  'boot_images': {'os/{}/boot'.format(arch): 'bzImage.*',
                                  'os/{}/ramdisks'.format(arch): 'usb-ramdisk*'}
                  },
        's3': {'patterns': ['Sourcefire_3D_Device_S3-{}{}-{}.iso'.format(version, feature, image),
                            'Cisco_Firepower_NGIPS_Appliance-{}{}-{}.iso'.format(version, feature, image)],
               'subdir': ['iso', 'iso/doNotRelease'],
               'boot_images': {'os/{}/boot'.format(arch): 'bzImage.*',
                               'os/{}/ramdisks'.format(arch): 'usb-ramdisk*'}
               },
        'kp': {'patterns': ['cisco-ftd-fp2k[\d.-]+[a-zA-Z]{3}', 'fxos-k8-fp2k-lfbff[\w.-]+[a-zA-Z]{3}',
                            'fxos-k8-lfbff[\w.-]+[a-zA-Z]{3}'],
               'subdir': ['installers', 'installers/doNotRelease'],
               },
        'ssp': {'patterns': ['cisco-ftd[\d.-]+[a-zA-Z]{3}.csp'],
                'subdir': ['installers', 'installers/doNotRelease'],
                }
    }
    return devices



import logging
import traceback
import sys
from collections import defaultdict
import time
from ats import aetest

KICK_EXTERNAL = False

try:
    from kick.file_servers.file_servers import *
except ImportError:
    KICK_EXTERNAL = True
    pass

from kick.device2.fxos.app import AppInstance
from kick.device2.fxos.app_handler import AppInstanceUtils
from kick.device2.fxos.interfaces.interface_handler import InterfaceHandler
from kick.device2.fxos.interfaces.physical_interface import PhysicalInterface
from kick.device2.fxos.interfaces.portchannel_interface import PortChannelInterface
from kick.device2.fxos.interfaces.portchannel_subinterface import PortChannelSubInterface
from kick.device2.fxos.interfaces.subinterface import SubInterface
from kick.device2.fxos.logical_device import LogicalDevice
from kick.device2.fxos.logical_device_handler import LogicalDeviceUtils
from kick.device2.fxos.management_bootstrap import ManagementBootstrap, ManagementClusterBootstrap
from kick.device2.fxos.cluster_bootstrap import ClusterBootstrap
from kick.device2.fxos.resource import ResourceProfile
from kick.device2.fxos.resource_profile_handler import ResourceProfileUtils

logger = logging.getLogger(__name__)


def prepare_bootstrap_info(logical_device_info):
    try:
        registration_key = logical_device_info['bootstrap_keys_secret'].get("registration_key")
    except:
        registration_key = None
    try:
        password = logical_device_info['bootstrap_keys_secret'].get("password")
    except:
        password = None
    try:
        firepower_manager_ip = logical_device_info['bootstrap_keys'].get("firepower_manager_ip")
    except:
        firepower_manager_ip = None
    try:
        firewall_mode = logical_device_info['bootstrap_keys'].get("firewall_mode")
    except:
        firewall_mode = None
    try:
        fqdn = logical_device_info['bootstrap_keys'].get("fqdn")
    except:
        fqdn = None
    try:
        dns_servers = logical_device_info['bootstrap_keys'].get("dns_servers")
    except:
        dns_servers = None
    try:
        search_domains = logical_device_info['bootstrap_keys'].get("search_domains")
    except:
        search_domains = None
    try:
        permit_expert_mode = logical_device_info['bootstrap_keys'].get("permit_expert_mode")
    except:
        permit_expert_mode = None
    try:
        management_type = logical_device_info['bootstrap_keys'].get('management_type')
    except:
        management_type = None
    return registration_key, password, firepower_manager_ip, firewall_mode, fqdn, dns_servers, search_domains, permit_expert_mode, management_type


def prepare_management_bootstrap(handle, logical_device_info):
    registration_key, password, firepower_manager_ip, firewall_mode, fqdn, dns_servers, search_domains, permit_expert_mode, management_type = prepare_bootstrap_info(
        logical_device_info)
    try:
        ipv4_ip = logical_device_info['ipv4'].get("ip")
    except:
        ipv4_ip = None
    try:
        ipv4_netmask = logical_device_info['ipv4'].get("netmask")
    except:
        ipv4_netmask = None
    try:
        ipv4_gateway = logical_device_info['ipv4'].get("gateway")
    except:
        ipv4_gateway = None
    try:
        ipv6_ip = logical_device_info['ipv6'].get("ip")
    except:
        ipv6_ip = None
    try:
        ipv6_prefix = logical_device_info['ipv6'].get("prefix")
    except:
        ipv6_prefix = None
    try:
        ipv6_gateway = logical_device_info['ipv6'].get("gateway")
    except:
        ipv6_gateway = None
    return ManagementBootstrap(handle=handle, ipv4_address=ipv4_ip,
                               ipv4_netmask=ipv4_netmask,
                               ipv4_gateway=ipv4_gateway,
                               ipv6_address=ipv6_ip,
                               ipv6_prefix=ipv6_prefix,
                               ipv6_gateway=ipv6_gateway, registration_key=registration_key, password=password,
                               firepower_manager_ip=firepower_manager_ip, firewall_mode=firewall_mode, fqdn=fqdn,
                               dns_servers=dns_servers, search_domains=search_domains,
                               permit_expert_mode=permit_expert_mode, management_type=management_type)


def prepare_intra_cluster_management_bootstrap(handle, logical_device_info):
    registration_key, password, firepower_manager_ip, firewall_mode, fqdn, dns_servers, search_domains, permit_expert_mode, management_type = prepare_bootstrap_info(
        logical_device_info)

    return ManagementClusterBootstrap(handle=handle, ip_config=logical_device_info['ip_config'],
                                      registration_key=registration_key, password=password,
                                      firepower_manager_ip=firepower_manager_ip, firewall_mode=firewall_mode,
                                      fqdn=fqdn, dns_servers=dns_servers, search_domains=search_domains,
                                      permit_expert_mode=permit_expert_mode, management_type=management_type)


def prepare_cluster_bootstrap(handle, logical_cluster_info):
    try:
        chassis_id = logical_cluster_info.get("chassis_id")
    except:
        chassis_id = None
    try:
        ccl_network = logical_cluster_info.get("ccl_network")
    except:
        ccl_network = None
    try:
        cluster_key = logical_cluster_info.get("cluster_key")
    except:
        cluster_key = None
    try:
        cluster_name = logical_cluster_info.get("cluster_name")
    except:
        cluster_name = None
    try:
        site_id = logical_cluster_info.get("site_id")
    except:
        site_id = None
    return ClusterBootstrap(handle=handle, chassis_id=chassis_id, ccl_network=ccl_network, cluster_key=cluster_key,
                            cluster_name=cluster_name, site_id=site_id)


class PackagePrerequisites(aetest.Testcase):
    handle = None
    # the path to the location of the install file e.g. 'Development', 'Feature/ASA_INTEG'
    branch_path = None
    version = None
    build = None
    site = None
    fxos_file = None
    version_build = None
    server_ip = None

    @aetest.test
    def init_params(self):
        pass

    @aetest.test
    def delete_old_ftd_images(self, steps):
        try:
            with steps.start("Delete old FTD images"):
                self.handle.execute('top')
                self.handle.execute('scope ssa')
                result = self.handle.execute('show app')
                ftd_images = result.splitlines()
                if len(ftd_images) > 1:
                    for image in ftd_images:
                        items = image.split()
                        if items[0] == 'ftd' and items[1] != self.version + '.' + self.build:
                            delete_command = "delete app ftd {}".format(items[1])
                            try:
                                self.handle.execute(delete_command)
                                output = self.handle.execute('commit-buffer')
                                if 'Error' in output:
                                    raise Exception("Found error in output - moving to next app")
                            except Exception:
                                self.handle.execute('discard-buffer')
                                pass
                self.handle.execute('show app')
        except Exception as e:
            logger.info("Unable to delete old FTD images e= {} {} {}".format(e, traceback.format_exc(),
                                                                             sys.exc_info()))
            pass

    @aetest.test
    def delete_old_fxos_images(self, steps):
        try:
            with steps.start("Delete old FXOS images"):
                self.handle.execute('scope firmware')
                result = self.handle.execute('show package')
                fxos_images = result.splitlines()
                if len(fxos_images) > 1:
                    for image in fxos_images:
                        items = image.split()
                        if 'fxos' in items[0] and items[0] != self.fxos_file:
                            delete_command = "delete package {}".format(items[0])
                            try:
                                self.handle.execute(delete_command)
                            except Exception:
                                pass
                time.sleep(5)
                self.handle.execute('show package')
        except Exception as e:
            logger.info("Unable to delete old FXOS images e= {} {} {}".format(e, traceback.format_exc(),
                                                                              sys.exc_info()))
            pass

    @aetest.test
    def download_fxos(self, steps):
        try:
            if self.fxos_file is not None:
                with steps.start("Prepare fxos file on PXE"):
                    self.server_ip = get_fxos_file(self.site, self.fxos_file)
        except Exception as e:
            self.failed(
                "Unable to download FXOS file on PXE server e= {} {} {}".format(e, traceback.format_exc(), sys.exc_info()),
                goto=["common_cleanup"])
        try:
            if self.fxos_file is not None:
                with steps.start("Download FXOS"):
                    fxos_url = 'scp://pxe@{}:{}/{}'.format(self.server_ip, pxe_dir['fxos_dir'], self.fxos_file)
                    self.handle.download_fxos(fxos_url=fxos_url, file_server_password=pxe_password)
        except Exception as e:
            self.failed("Unable to download FXOS file on chassis e= {} {} {}".format(e, traceback.format_exc(),
                                                                                     sys.exc_info()),
                        goto=["common_cleanup"])

    @aetest.test
    def prepare_files(self, steps):
        try:
            with steps.start("Prepare files on PXE"):
                self.server_ip, tftp_prefix, self.scp_prefix, self.files = prepare_installation_files(self.site, 'Ssp',
                                                                                                      self.branch_path,
                                                                                                      self.version_build,
                                                                                                      fxos_file=self.fxos_file)
        except Exception as e:
            self.failed(
                "Unable to download files on PXE server e= {} {} {}".format(e, traceback.format_exc(), sys.exc_info()),
                goto=["common_cleanup"])

    @aetest.test
    def download_csp(self, steps):
        try:
            with steps.start("Download CSP"):
                csp_file = [file for file in self.files if file.endswith('.csp')][0]
                scp_csp_link = "scp://pxe@{}:/{}/{}".format(self.server_ip, self.scp_prefix, csp_file)
                self.handle.download_csp(csp_url=scp_csp_link, file_server_password=pxe_password)
        except Exception as e:
            self.failed("Unable to download CSP file on chassis e= {} {} {}".format(e, traceback.format_exc(),
                                                                                    sys.exc_info()),
                        goto=["common_cleanup"])


class ConfigurationClean(aetest.Testcase):
    handle = None
    configuration = None
    slots = [1, 2, 3]

    @aetest.test
    def init_params(self):
        pass

    @aetest.test
    def clean_logical_device(self):
        self.handle.execute('discard-buffer')
        try:
            logical_device_handler = LogicalDeviceUtils(self.handle)
            logical_device_handler.delete_all_ld()
        except Exception as e:
            self.failed(
                "Unable to clean App Instances e= {} {} {}".format(e, traceback.format_exc(), sys.exc_info()),
                goto=["common_cleanup"])

    @aetest.test
    def clean_app_instance(self):
        try:
            for slot in self.slots:
                app_handler = AppInstanceUtils(self.handle, slot)
                apps = app_handler.get_app_instances_in_slot()
                for app in apps:
                    app_instance = AppInstance(handle=self.handle, app_name=app['App Name'], slot=slot,
                                               identifier=app['Identifier'])
                    app_instance.delete()
        except Exception as e:
            self.failed(
                "Unable to clean App Instances e= {} {} {}".format(e, traceback.format_exc(), sys.exc_info()),
                goto=["common_cleanup"])

    @aetest.test
    def resource_profile(self):
        try:
            resource_handler = ResourceProfileUtils(self.handle)
            resource_handler.delete_all()
        except Exception as e:
            self.failed(
                "Unable to clean App Instances e= {} {} {}".format(e, traceback.format_exc(), sys.exc_info()),
                goto=["common_cleanup"])


class InterfaceCleanUtils(aetest.Testcase):
    handle = None

    @aetest.test
    def init_params(self):
        pass

    @aetest.test
    def clean_portchannel(self):
        try:
            interface_handler = InterfaceHandler(self.handle)
            interface_handler.delete_all_port_channel()
        except Exception as e:
            self.failed(
                "Unable to delete all portchannel e= {} {} {}".format(e, traceback.format_exc(), sys.exc_info()),
                goto=["common_cleanup"])

    @aetest.test
    def clean_interfaces(self):
        try:
            interface_handler = InterfaceHandler(self.handle)
            interface_handler.reset_all_physical_interfaces()
        except Exception as e:
            self.failed(
                "Unable to clean physical interface e= {} {} {}".format(e, traceback.format_exc(), sys.exc_info()),
                goto=["common_cleanup"])


class ConfigureInterfacesUtils(aetest.Testcase):
    handle = None
    interfaces = None

    @aetest.test
    def init_params(self):
        pass

    def configure_sub_interface(self, hardware, subinterface_input):
        try:
            for interface in subinterface_input:
                subinterface_obj = SubInterface(handle=self.handle, hardware=hardware,
                                                sub_interface_id=interface.get("id"))
                try:
                    port_type = interface.get("subtype")
                except:
                    port_type = 'data'
                subinterface_obj.configure(vlan_id=interface.get("vlan_id"), port_type=port_type)
        except Exception as e:
            raise e

    def configure_port_channel_sub_interface(self, hardware, subinterface_input):
        try:
            for interface in subinterface_input:
                subinterface_obj = PortChannelSubInterface(handle=self.handle, hardware=hardware,
                                                           sub_interface_id=interface.get("id"))
                try:
                    port_type = interface.get("subtype")
                except:
                    port_type = 'data'
                subinterface_obj.configure(vlan_id=interface.get("vlan_id"), port_type=port_type)
        except Exception as e:
            raise e

    @aetest.test
    def configure_interfaces(self, steps):
        portchannel = []
        interface_handler = InterfaceHandler(self.handle)
        with steps.start('Create physical interfaces and subinterfaces') as step:
            try:
                for interface in self.interfaces:
                    if getattr(self.interfaces[interface], "hardware", None):
                        with step.start('Create physical interface {}'.format(self.interfaces[interface].hardware)):
                            intf_object = PhysicalInterface(self.handle, self.interfaces[interface].hardware)
                            intf_object.configure(port_type=self.interfaces[interface].subtype, enabled=True)

                            if getattr(self.interfaces[interface], "subinterfaces", None):
                                self.configure_sub_interface(self.interfaces[interface].hardware,
                                                             self.interfaces[interface].subinterfaces)
                    else:
                        portchannel.append(interface)
            except Exception as e:
                self.failed(
                    "Unable to determine interface type e= {} {} {}".format(e, traceback.format_exc(), sys.exc_info()),
                    goto=["common_cleanup"])
        with steps.start('Create portchannel interfaces and portchannel subinterfaces') as step:
            try:
                for interface in portchannel:
                    name = "Port-channel{}".format(self.interfaces[interface].id)
                    with step.start('Create portchannel interface {}'.format(name)):
                        intf_object = PortChannelInterface(self.handle, name)
                        member_interfaces = None
                        if hasattr(self.interfaces[interface], 'member_ports'):
                            member_interfaces = []
                            for member in self.interfaces[interface].member_ports:
                                member_interfaces.append(
                                    interface_handler.instantiate_interface_based_on_name(name=member))
                        intf_object.configure(port_type=self.interfaces[interface].subtype,
                                              member_ports=member_interfaces, enabled=True)
                        if getattr(self.interfaces[interface], "subinterfaces", None):
                            self.configure_port_channel_sub_interface(name, self.interfaces[interface].subinterfaces)
            except Exception as e:
                self.failed(
                    "Unable to determine interface type e= {} {} {}".format(e, traceback.format_exc(), sys.exc_info()),
                    goto=["common_cleanup"])


class ConfigureMultiInstanceUtils(aetest.Testcase):
    handle = None
    configuration = None
    apps_to_create = []
    app_list = []

    logical_devices = {}

    @aetest.test
    def init_params(self):
        pass

    @aetest.test
    def build_resource_profile(self):
        profiles = self.configuration.get('resource_profiles', [])
        for resource_profile in profiles:
            try:
                rp = ResourceProfile(handle=self.handle, name=profiles[resource_profile].get("name"))
                rp.configure(cpu=profiles[resource_profile].get("cpu_core_count"))
            except Exception as e:
                self.failed(
                    "Unable to create Resource Profile e= {} {} {}".format(e, traceback.format_exc(), sys.exc_info()),
                    goto=["common_cleanup"])

    def add_unique_key_in_logical_device(self, key, logical_device):
        # only one of the apps in a cluster can contain some keys
        if key in logical_device:
            if key in self.logical_devices[logical_device['name']]:
                raise Exception("This logical device {} has already a {}".format(logical_device['name'], key))
            self.logical_devices[logical_device['name']][key] = logical_device[key]

    def add_key_if_values_match(self, key, logical_device):
        # some keys in a cluster can be in more than one app, but they need to have the same value
        if key in logical_device:
            if key in self.logical_devices[logical_device['name']] and self.logical_devices[logical_device['name']][
                key] != logical_device[key]:
                raise Exception(
                    "This logical device {} has already a {} with another value".format(logical_device['name'], key))
            self.logical_devices[logical_device['name']][key] = logical_device[key]

    def add_ip_for_slot(self, slot, logical_device):
        self.logical_devices[logical_device['name']]['ip_config'][slot] = {}
        if 'ipv4' in logical_device:
            self.logical_devices[logical_device['name']]['ip_config'][slot]['ipv4'] = logical_device['ipv4']
        if 'ipv6' in logical_device:
            self.logical_devices[logical_device['name']]['ip_config'][slot]['ipv6'] = logical_device['ipv6']

    def prepare_logical_device_data(self, logical_device, app_instance, slot):
        # creates a dictionary for logical devices having the logical device name as key
        if logical_device['name'] not in self.logical_devices:
            self.logical_devices[logical_device['name']] = logical_device
            self.logical_devices[logical_device['name']]['app'] = app_instance
            self.logical_devices[logical_device['name']]['ip_config'] = {}
            self.add_ip_for_slot(slot, logical_device)
        else:
            if app_instance.app_name != self.logical_devices[logical_device['name']]['app'].app_name:
                raise Exception("This logical device {} has associated this app name {} and you want to set {}".format(
                    logical_device['name'], self.logical_devices[logical_device['name']]['app'].app_name,
                    app_instance.app_name))
            self.add_key_if_values_match('logical_slot', logical_device)
            self.add_key_if_values_match('device_mode', logical_device)
            self.add_unique_key_in_logical_device('bootstrap_cluster', logical_device)
            self.add_unique_key_in_logical_device('external_port_links', logical_device)
            self.add_unique_key_in_logical_device('bootstrap_keys', logical_device)
            self.add_unique_key_in_logical_device('bootstrap_keys_secret', logical_device)
            if slot in self.logical_devices[logical_device['name']]['ip_config']:
                raise Exception("We already have an app on slot {}".format(slot))
            self.add_ip_for_slot(slot, logical_device)

    @aetest.test
    def build_app_instances(self, steps):
        instances = self.configuration['applications']
        for i in instances:
            if (len(self.apps_to_create) > 0 and instances[i].get(
                    "application_identifier") in self.apps_to_create) or len(self.apps_to_create) == 0:
                with steps.start('Create App Instance {}'.format(instances[i].get("application_identifier"))):
                    try:
                        app_instance = AppInstance(handle=self.handle, app_name=instances[i].get("application_name"),
                                                   slot=instances[i].get("slot"),
                                                   identifier=instances[i].get("application_identifier"))
                        hw_crypto_state = None
                        if 'hw-crypto' in instances[i]:
                            if 'admin-state' in instances[i]['hw-crypto']:
                                    hw_crypto_state = instances[i]['hw-crypto'].get("admin-state")
                            else:

                                self.failed("hw-crypto admin-state was not configured correct in testbed",goto=["common_cleanup"])
                        app_instance.create(deploy_type=instances[i].get("deploy_type"),
                                            version=instances[i].get("startup_version"),
                                            resource_profile=instances[i].get("resource_profile"),hw_crypto_state = hw_crypto_state)
                        self.app_list.append(app_instance)
                        self.prepare_logical_device_data(instances[i]['logical_device'], app_instance,
                                                         instances[i].get("slot"))
                    except Exception as e:
                        self.failed("Unable to create App Instance with name={} identifier={} e= {} {} {}".format(
                            instances[i].get("application_name"), instances[i].get("application_identifier"), e,
                            traceback.format_exc(),
                            sys.exc_info()),
                            goto=["common_cleanup"])

    @aetest.test
    def build_logical_devices(self, steps):
        interface_handler = InterfaceHandler(self.handle)
        for i in self.logical_devices:
            logical_device_info = self.logical_devices[i]
            with steps.start('Create Logical Device with name {}'.format(logical_device_info.get("name"))):
                try:
                    cluster_bootstrap = None
                    if 'bootstrap_cluster' in logical_device_info:
                        logical_cluster_info = logical_device_info.get("bootstrap_cluster")
                        cluster_bootstrap = prepare_cluster_bootstrap(self.handle, logical_cluster_info)
                    if len(logical_device_info['ip_config'].keys()) > 1:
                        mgmt_bootstrap = prepare_intra_cluster_management_bootstrap(self.handle, logical_device_info)
                    else:
                        mgmt_bootstrap = prepare_management_bootstrap(self.handle, logical_device_info)
                    interface_objects = []
                    if 'external_port_links' in logical_device_info:
                        for interface in logical_device_info.get("external_port_links"):
                            interface_obj = interface_handler.instantiate_interface_based_on_name(name=interface)
                            interface_objects.append(interface_obj)
                    if 'logical_slot' in logical_device_info:
                        slot = '\"' + logical_device_info.get("logical_slot") + '\"'
                    else:
                        slot = logical_device_info.get("app").slot
                    logical_device = LogicalDevice(handle=self.handle, name=logical_device_info.get("name"),
                                                   app=logical_device_info.get("app"),
                                                   slot=slot,
                                                   type=logical_device_info.get("device_mode"))
                    logical_device.create(interfaces=interface_objects, mgmt_bootstrap=mgmt_bootstrap,
                                          cluster_bootstrap=cluster_bootstrap)
                except Exception as e:
                    self.failed(
                        "Unable to create logical device e= {} {} {}".format(e, traceback.format_exc(),
                                                                             sys.exc_info()),
                        goto=["common_cleanup"])

    def handle_cluster_exception(self, app, problem_apps):
        try:
            logger.info(
                "Handling exception for cluster - the product requires to provide information for all 3 blades, but cluster can be formed with only 2 blades.")
            logger.info("Check if status for the app is 'Not Available'")
            app_details = app.get_app_instance_detail()
            if app_details['Oper State'] != "Not Available":
                logger.error("We are handling cluster exception only if the blade is missing.")
                return False
            logger.info("Status is 'Not Available'. Moving to next check.")
            logger.info("Checking if app is part of cluster.")
            device_mode = None
            instances = self.configuration['applications']
            for i in instances:
                if instances[i].get("application_identifier") == app.identifier and instances[i].get("slot") == app.slot:
                    device_mode = instances[i]["logical_device"]["device_mode"]
                    break
            if not device_mode or device_mode != 'clustered':
                logger.error("This device mode is {}. We are handling exception only for clustered.".format(device_mode))
                return False
            logger.info("App is part of a cluster. Moving to next check.")
            logger.info("Checking if the other apps on same cluster are online.")
            count = 0
            for problem_app in problem_apps:
                if problem_app.identifier == app.identifier:
                    count += 1
            if count > 1:
                logger.error("There are {} with same name {} in a bad state. Cannot form cluster.".format(count, app.identifier))
                return False
            logger.info("The other apps are online.")
        except Exception as e:
            logger.error("Failed while checking if there is an exception for the app. e={}".format(e))
            return False
        return True

    @aetest.test
    def wait_until_online(self):
        failed_apps = []
        to_check_apps = self.app_list
        step = 0
        while step < 180 and len(to_check_apps) > 0:
            logger.info('Checking if enable - {} try'.format(step))
            time.sleep(30)
            not_completed_apps = []
            for app in to_check_apps:
                try:
                    if not app.check_status():
                        not_completed_apps.append(app)
                except:
                    failed_apps.append(app)
            to_check_apps = not_completed_apps
            logger.info('Number of failed apps: {}'.format(len(failed_apps)))
            logger.info('Waiting for {} apps'.format(len(to_check_apps)))
            step += 1
        if len(failed_apps) > 0 or len(to_check_apps) > 0:
            catastrophic_failure = False
            if len(failed_apps) > 0:
                logger.error("We have failed apps. Checking if they are known exceptions:")
                summary = ''
                for app in failed_apps:
                    logger.error("\t Failed app {} on slot {}".format(app.identifier, app.slot))
                    problem_apps = []
                    problem_apps.extend(failed_apps)
                    problem_apps.extend(to_check_apps)
                    if not self.handle_cluster_exception(app, problem_apps):
                        catastrophic_failure = True
                        summary = "{}\n\t app {} on slot {}".format(summary, app.identifier, app.slot)
                if summary != '':
                    logger.error("We have failed apps: {}".format(summary))
            if len(to_check_apps) > 0:
                logger.error("We have apps which aren't Online yet: ")
                catastrophic_failure = True
                for app in to_check_apps:
                    logger.error("\t app {} on slot {}".format(app.identifier, app.slot))
            if catastrophic_failure:
                self.failed("Unable to start all logical devices", goto=["common_cleanup"])
            else:
                self.passx("Some apps were not started, but these are known exceptions")



from unicon.eal.dialogs import Statement


class AsaStatements:
    def __init__(self, patterns):
        self.disable_to_enable_statement = Statement(
        	pattern=patterns.prompt.password_prompt,
            action=lambda spawn, password: spawn.sendline(password),
            args={'password': patterns.enable_password},
            loop_continue=True, continue_timer=True
          )

        self.set_enable_pwd_statement = Statement(
        	pattern=patterns.prompt.set_enable_pwd_prompt,
        	action=lambda spawn, password: spawn.sendline(password),
        	args={'password': patterns.enable_password},
        	loop_continue=True, continue_timer=True
        )

        self.repeat_enable_pwd_statement = Statement(
        	pattern=patterns.prompt.repeat_enable_pwd_prompt,
        	action=lambda spawn, password: spawn.sendline(password),
        	args={'password': patterns.enable_password},
        	loop_continue=True, continue_timer=True
        )



"""Perform power-bar options on a device."""
import logging
import re
import telnetlib
import time

try:
    from kick.misc.convert_bytes import string_to_bytes, bytes_to_string
except ImportError:
    from kick.miscellaneous.convert import string_to_bytes, bytes_to_string

LOGGER = logging.getLogger(__name__)
HANDLER = logging.StreamHandler()
FORMATTER = logging.Formatter('%(asctime)s : %(levelname)s : %(name)s : %(message)s')
HANDLER.setFormatter(FORMATTER)
LOGGER.addHandler(HANDLER)


def power_bar(
              power_server,
              port,
              action='status',
              user='admn',
              pwd='admn'):
    """Telnet to power-bar and perform the specified action

    :param power_server: name or IP Address of power-bar
    :param port: port of the device to perform power action
    :param action: status, on, off, or reboot
    :param user: power-bar credential
    :param pwd: power-bar credential
    :return:
 
    """

    actions = ['status', 'on', 'off', 'reboot']
    LOGGER.info('power_bar %s port=%s action=%s' % (power_server, port, action))
    if not action.lower() in actions:
        raise ValueError('action should be one of %s' % str(actions))

    # defined common patterns for the different versions of PDU
    pattern = r'.*\.?%s +[^ ]+ +([^ ]+).*' % port if action.lower() == 'status' \
        else r'.*(Command successful).*'

    try:
        # match both 'Switched PDU:' and 'Switched CDU:' menu selection
        prompt = [string_to_bytes("Switched .*:")]
        session = telnetlib.Telnet(power_server)
        session.read_until(string_to_bytes("Username:"))
        session.write(string_to_bytes(user + '\n'))
        session.read_until(string_to_bytes("Password:"))
        session.write(string_to_bytes(pwd + '\n'))
        session.expect(prompt)
        session.write(string_to_bytes('%s .%s\n' % (action, port)))
        # expect returns a tuple, where the last element is
        # the text read up till and including the match
        result = session.expect(prompt)[2]
        for line in result.splitlines():
            LOGGER.debug(line)
        return re.search(pattern, bytes_to_string(result), re.IGNORECASE).group(1)
    finally:
        try:
            session.close()
        except:
            pass


def power_cycle_all_ports(power_bar_server, power_bar_port, power_bar_user, power_bar_pwd):
    """Powers off and then powers on all given ports.

    :param power_bar_server: comma-separated string of IP addresses of the PDU's
    :param power_bar_port: comma-separated string of power port on the PDU's
    :param power_bar_user: comma-separated usernames for power bar servers
    :param power_bar_pwd:  comma-separated passwords for power bar servers
    :return: True if all ports were powered off and on successfully, False otherwise
    """

    result = True
    power_bar_servers = [server.strip() for server in power_bar_server.split(',')]
    power_bar_ports = [port.strip() for port in power_bar_port.split(',')]
    power_bar_users = [user.strip() for user in power_bar_user.split(',')]
    power_bar_pwds = [pwd.strip() for pwd in power_bar_pwd.split(',')]
    LOGGER.info('->Power off all power ports')
    for server, port, user, pwd in zip(power_bar_servers, power_bar_ports, power_bar_users, power_bar_pwds):
        LOGGER.info('->Power off {} {}'.format(server, port))
        result = result and power_bar(server, port, action='off', user=user, pwd=pwd)
        LOGGER.info('->Done ')
        time.sleep(10)
    LOGGER.info("Sleeping for 60 secs..")
    time.sleep(60)
    LOGGER.info('->Power on all power ports')
    for server, port, user, pwd in zip(power_bar_servers, power_bar_ports, power_bar_users, power_bar_pwds):
        LOGGER.info('->Power on {} {}'.format(server, port))
        result = result and power_bar(server, port, action='on', user=user, pwd=pwd)
        LOGGER.info('->Done')
        time.sleep(10)

    return result
